# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from dataclasses import dataclass, field
from typing import Optional
from odxtools.utils import read_description_from_odx


@dataclass()
class Audience:
    enabled_audience_refs: list = field(default_factory=list)
    disabled_audience_refs: list = field(default_factory=list)
    is_supplier: bool = True
    is_development: bool = True
    is_manufacturing: bool = True
    is_aftersales: bool = True
    is_aftermarket: bool = True

    @property
    def enabled_audiences(self):
        return self._enabled_audiences

    @property
    def disabled_audiences(self):
        return self._disabled_audiences

    def _resolve_references(self, id_lookup):
        self._enabled_audiences = [id_lookup[ref]
                                   for ref in self.enabled_audience_refs]
        self._disabled_audiences = [id_lookup[ref]
                                    for ref in self.disabled_audience_refs]


@dataclass()
class AdditionalAudience:
    """
    Corresponds to ADDITIONAL-AUDIENCE.
    """
    id: str
    short_name: str
    long_name: Optional[str] = None
    description: Optional[str] = None


def read_audience_from_odx(et_element):
    enabled_audience_refs = [ref.get(
        "ID-REF") for ref in et_element.iterfind("ENABLED-AUDIENCE-REFS/ENABLED-AUDIENCE-REF")]
    disabled_audience_refs = [ref.get(
        "ID-REF") for ref in et_element.iterfind("DISABLED-AUDIENCE-REFS/DISABLED-AUDIENCE-REF")]
    is_supplier = et_element.get(
        "IS-SUPPLIER") == 'true' if et_element.get("IS-SUPPLIER") else True
    is_development = et_element.get(
        "IS-DEVELOPMENT") == 'true' if et_element.get("IS-DEVELOPMENT") else True
    is_manufacturing = et_element.get(
        "IS-MANUFACTURING") == 'true' if et_element.get("IS-MANUFACTURING") else True
    is_aftersales = et_element.get(
        "IS-AFTERSALES") == 'true' if et_element.get("IS-AFTERSALES") else True
    is_aftermarket = et_element.get(
        "IS-AFTERMARKET") == 'true' if et_element.get("IS-AFTERMARKET") else True

    return Audience(enabled_audience_refs=enabled_audience_refs,
                    disabled_audience_refs=disabled_audience_refs,
                    is_supplier=is_supplier,
                    is_development=is_development,
                    is_manufacturing=is_manufacturing,
                    is_aftersales=is_aftersales,
                    is_aftermarket=is_aftermarket)


def read_additional_audience_from_odx(et_element):
    short_name = et_element.find("SHORT-NAME").text
    id = et_element.get("ID")

    long_name = et_element.find(
        "LONG-NAME").text if et_element.find("LONG-NAME") is not None else None
    description = read_description_from_odx(et_element.find("DESC"))

    return AdditionalAudience(id=id,
                              short_name=short_name,
                              long_name=long_name,
                              description=description)



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH


from ..dataobjectproperty import DataObjectProperty
from ..encodestate import EncodeState

from .parameterwithdop import ParameterWithDOP


class ValueParameter(ParameterWithDOP):
    def __init__(self,
                 short_name,
                 physical_default_value=None,
                 dop_ref=None,
                 dop_snref=None,
                 **kwargs):
        super().__init__(short_name,
                         parameter_type="VALUE",
                         dop_ref=dop_ref,
                         dop_snref=dop_snref,
                         **kwargs)
        # _physical_default_value is a string. Conversion to actual type must happen after parsing
        self._physical_default_value = physical_default_value

    @property
    def physical_default_value(self):
        if self._physical_default_value is None:
            return None
        else:
            return self.dop.physical_type.base_data_type.from_string(self._physical_default_value)

    def is_required(self):
        return True if self.physical_default_value is None else False

    def is_optional(self):
        return True if self._physical_default_value is not None else False

    def get_coded_value(self, physical_value=None):
        if physical_value is not None:
            return self.dop.convert_physical_to_internal(physical_value)
        else:
            return self.dop.convert_physical_to_internal(self.physical_default_value)

    def get_coded_value_as_bytes(self, encode_state: EncodeState):
        physical_value = encode_state.parameter_values.get(self.short_name,
                                                           self.physical_default_value)
        if physical_value is None:
            raise TypeError(f"A value for parameter '{self.short_name}' must be specified"
                            f" as the parameter does not exhibit a default.")
        assert self.dop is not None, f"Param {self.short_name} does not have a DOP. Maybe resolving references failed?"
        return self.dop.convert_physical_to_bytes(physical_value, encode_state=encode_state, bit_position=self.bit_position)

    def get_valid_physical_values(self):
        if isinstance(self.dop, DataObjectProperty):
            return self.dop.get_valid_physical_values()

    def __repr__(self):
        repr_str = f"ValueParameter(short_name='{self.short_name}'"
        if self.long_name is not None:
            repr_str += f", long_name='{self.long_name}'"
        if self.byte_position is not None:
            repr_str += f", byte_position='{self.byte_position}'"
        if self.bit_position:
            repr_str += f", bit_position='{self.bit_position}'"
        if self.semantic is not None:
            repr_str += f", semantic='{self.semantic}'"
        if self.physical_default_value is not None:
            repr_str += f", physical_default_value='{self.physical_default_value}'"
        if self.dop_ref is not None:
            repr_str += f", dop_ref='{self.dop_ref}'"
        if self.dop_snref is not None:
            repr_str += f", dop_snref='{self.dop_snref}'"
        if self.description is not None:
            repr_str += f", description='{' '.join(self.description.split())}'"
        return repr_str + ")"



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from .database import Database
from .globals import logger

def load_odx_d_file(odx_d_file_name: str, enable_candela_workarounds: bool = True):
    container = Database(odx_d_file_name=odx_d_file_name,
                         enable_candela_workarounds=enable_candela_workarounds)
    logger.info(f"--- --- --- Done with parsing --- --- ---")
    return container



#! /usr/bin/python3
#
# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

import bitstruct
import can
import asyncio
import re
import sys
from enum import IntEnum

class IsoTp(IntEnum):
    # Frame types
    FRAME_TYPE_SINGLE = 0
    FRAME_TYPE_FIRST = 1
    FRAME_TYPE_CONSECUTIVE = 2
    FRAME_TYPE_FLOW_CONTROL = 3

    # Flow control flag values
    FLOW_CONTROL_CONTINUE = 0
    FLOW_CONTROL_WAIT = 1
    FLOW_CONTROL_ABORT = 2

class IsoTpStateMachine:
    can_normal_frame_re = re.compile("([a-zA-Z0-9_-]*) *([0-9A-Fa-f ]*) *\[[0-9]*\] *([ 0-9A-Fa-f]*)")
    can_log_frame_re = re.compile("\([0-9.]*\) *([a-zA-Z0-9_-]*) ([0-9A-Fa-f]*)#([0-9A-Fa-f]*)")

    def __init__(self, can_rx_ids):
        if isinstance(can_rx_ids, int):
            can_rx_ids = [can_rx_ids]

        self._can_rx_ids = can_rx_ids
        assert(isinstance(self._can_rx_ids, list))

        self._telegram_specified_len = [0]*len(can_rx_ids)
        self._telegram_data = [None]*len(can_rx_ids)
        self._telegram_last_rx_fragment_idx = [0]*len(can_rx_ids)

    def decode_rx_frame(self, rx_id, data, do_yield=False):
        """Handle the ISO-TP state transitions caused by a CAN frame.

        E.g., add some data to a telegram, etc.
        """
        try:
            telegram_idx = self._can_rx_ids.index(rx_id)
        except ValueError:
            return False # unknown CAN ID

        # decode the isotp segment
        result = True
        frame_type, = bitstruct.unpack("u4", data)
        telegram_len = None
        if frame_type == IsoTp.FRAME_TYPE_SINGLE:
            frame_type, telegram_len = bitstruct.unpack("u4u4", data)

            self.on_single_frame(telegram_idx, data[1:1+telegram_len])
            self.on_telegram_complete(telegram_idx, data[1:1+telegram_len])

            if do_yield:
                yield (rx_id, data[1:1+telegram_len])

        elif frame_type == IsoTp.FRAME_TYPE_FIRST:
            frame_type, telegram_len = bitstruct.unpack("u4u12", data)

            self._telegram_specified_len[telegram_idx] = telegram_len
            self._telegram_data[telegram_idx] = data[2:]
            self._telegram_last_rx_fragment_idx[telegram_idx] = 0

            self.on_first_frame(telegram_idx, data)

        elif frame_type == IsoTp.FRAME_TYPE_CONSECUTIVE:
            frame_type, rx_segment_idx = bitstruct.unpack("u4u4", data)

            expected_segment_idx = (self._telegram_last_rx_fragment_idx[telegram_idx] + 1)%16

            if expected_segment_idx == rx_segment_idx:
                self._telegram_last_rx_fragment_idx[telegram_idx] = rx_segment_idx
                self._telegram_data[telegram_idx] += data[1:]

                n = self._telegram_specified_len[telegram_idx]
                if len(self._telegram_data[telegram_idx]) > n:
                    # can frames can include padding, i.e. the length
                    # of the telegram payload is not necessarily a
                    # multiple of the segment payloads
                    self._telegram_data[telegram_idx] = \
                        self._telegram_data[telegram_idx][:n]
            else:
                result = False

            self.on_consecutive_frame(telegram_idx,
                                      rx_segment_idx,
                                      data[1:])

            if expected_segment_idx != rx_segment_idx:
                self.on_sequence_error(telegram_idx,
                                       expected_segment_idx,
                                       rx_segment_idx)
            elif len(self._telegram_data[telegram_idx]) == n:
                self.on_telegram_complete(telegram_idx, self._telegram_data[telegram_idx])
                if do_yield:
                    yield (rx_id, self._telegram_data[telegram_idx])

        elif frame_type == IsoTp.FRAME_TYPE_FLOW_CONTROL:
            frame_type, flow_control_flag = bitstruct.unpack("u4u4", data)
            self.on_flow_control_frame(telegram_idx, flow_control_flag)
        else:
            result = False
            self.on_frame_type_error(telegram_idx, frame_type)

        return result

    async def read_telegrams(self, bus):
        """This is equivalent to the :py:meth:`file.readlines()` method, but
        it yields ISO-TP telegrams instead of lines.

        The  yielded telegrams are (can_id, payload_data) tuples.

        :param bus: Input file or socket of can bus to read the can frames
        """

        if hasattr(bus, "set_filters"):
            # this is a bit of a hack: passing any object which
            # exhibits a set_filters() method is assumed to be a can
            # bus object.

            # creat an "on receive" event for the can bus
            rx_event = asyncio.Event()
            loop = asyncio.get_running_loop()
            loop.add_reader(bus, rx_event.set)

            while True:
                await rx_event.wait()

                msg = bus.recv()
                for tmp in self.decode_rx_frame(msg.arbitration_id,
                                                msg.data,
                                                do_yield=True):
                    yield tmp
        else:
            # input is a file

            while bus:
                cur_line = bus.readline()
                if cur_line == "":
                    return

                if m := self.can_normal_frame_re.match(cur_line.strip()):
                    frame_interface = m.group(1)
                    frame_id = int(m.group(2), 16)

                    frame_data_formatted = m.group(3).strip()
                    frame_data = bytearray(map(lambda x: int(x, 16), frame_data_formatted.split(" ")))

                    for tmp in self.decode_rx_frame(frame_id, frame_data, do_yield=True):
                        yield tmp

                elif m := self.can_log_frame_re.match(cur_line.strip()):
                    frame_interface = m.group(2)
                    frame_id = int(m.group(2), 16)

                    frame_data_formatted = m.group(3).strip()
                    frame_data_list = [ frame_data_formatted[i:i+2] for i in range(0, len(frame_data_formatted), 2) ]
                    frame_data = bytearray(map(lambda x: int(x, 16), frame_data_list))

                    for tmp in self.decode_rx_frame(frame_id, frame_data, do_yield=True):
                        yield tmp

                else:
                    print(f"Warning: unrecognized frame format: '{cur_line.strip()}'", file=sys.stderr)

    def can_rx_id(self, telegram_idx):

        """Given a Telegram index, returns the CAN ID for receiving data.

        :raises IndexError: The telegram index is invalid.
        """
        return self._can_rx_ids[telegram_idx]

    def telegram_data(self, telegram_idx):
        """Given a Telegram index, returns the data received for this telegram
        so far.

        :raises IndexError: The telegram index is invalid.
        """
        return self._telegram_data[telegram_idx]

    ##############
    # Callbacks
    ##############
    def on_single_frame(self, telegram_idx, frame_payload):
        """Method called when an ISO-TP message of type "single frame" has been received"""
        pass

    def on_first_frame(self, telegram_idx, frame_payload):
        """Method called when an ISO-TP message of type "first frame" has been received"""
        pass

    def on_consecutive_frame(self, telegram_idx, segment_idx, frame_payload):
        """Method called when an ISO-TP message of type "consecutive frame" has been received"""
        pass

    def on_flow_control_frame(self, telegram_idx, flow_control_flag):
        """Method called when an ISO-TP message of type "flow control frame" has been received"""
        pass

    def on_sequence_error(self, telegram_idx, expected_idx, rx_idx):
        """Method called when a frame with an unexpected sequence index has been received"""
        pass

    def on_frame_type_error(self, telegram_idx, frame_type):
        """Method called when a frame exhibiting an unknown frame type has been received"""
        pass

    def on_telegram_complete(self, telegram_idx, telegram_payload):
        """Method called when an ISO-TP telegram has been fully received"""
        pass


class IsoTpActiveDecoder(IsoTpStateMachine):
    """This class is equivalent to IsoTpStateMachine, but it sends out
    ISO-TP flow control messages (acknowledgements) to allow to
    receive ISO-TP messages actively instead of just snooping in on
    other people's conversations."""

    def __init__(self,
                 can_bus,
                 can_rx_ids,
                 can_tx_ids,
                 padding_size=0,
                 padding_value=0xaa):
        self._can_bus = can_bus

        if isinstance(can_tx_ids, int):
            can_tx_ids = [can_tx_ids]

        self._can_tx_ids = can_tx_ids
        self._padding_size = padding_size
        self._padding_value = padding_value

        super().__init__(can_rx_ids)

        assert(isinstance(self._can_tx_ids, list))
        assert(len(self._can_rx_ids) == len(self._can_tx_ids))
        assert(set(self._can_rx_ids).isdisjoint(set(self._can_tx_ids))) # correct?

        self._block_size = [None]*len(self._can_rx_ids)
        self._frames_received = [None]*len(self._can_rx_ids)

    def can_tx_id(self, telegram_idx):
        """Given a Telegram index, returns the CAN ID for sending data.

        :raises TypeError: No transmission IDs specified (e.g.,
                           because we are passively decoding)
        :raises IndexError: The telegram index is invalid.
        """
        return self._can_tx_ids[telegram_idx]

    def on_single_frame(self, telegram_idx, frame_payload):
        # send ACK
        rx_id = self.can_rx_id(telegram_idx)
        tx_id = self.can_tx_id(telegram_idx)
        block_size = 0xff
        min_separation_time = 0 # ms
        fc_payload = bitstruct.pack("u4u4u8u8",
                                    IsoTp.FRAME_TYPE_FLOW_CONTROL,
                                    IsoTp.FLOW_CONTROL_CONTINUE,
                                    block_size, # does not matter here?!
                                    min_separation_time)

        self._send_can_message(tx_id, fc_payload)
        self._frames_received[telegram_idx] = None

        super().on_first_frame(telegram_idx, frame_payload)

    def on_first_frame(self, telegram_idx, frame_payload):
        # send ACK
        rx_id = self.can_rx_id(telegram_idx)
        tx_id = self.can_tx_id(telegram_idx)
        block_size = 0xff # default value, can be overwritten later
        min_separation_time = 0 # ms
        fc_payload = bitstruct.pack("u4u4u8u8",
                                    IsoTp.FRAME_TYPE_FLOW_CONTROL,
                                    IsoTp.FLOW_CONTROL_CONTINUE,
                                    block_size,
                                    min_separation_time)

        self._send_can_message(tx_id, fc_payload)

        self._block_size[telegram_idx] = block_size
        # TODO: find out if the first frame counts into the current block...
        self._frames_received[telegram_idx] = 0

        super().on_first_frame(telegram_idx, frame_payload)

    def on_consecutive_frame(self, telegram_idx, segment_idx, frame_payload):
        self._frames_received[telegram_idx] += 1

        # send new ACK if necessary
        block_size = self._block_size[telegram_idx]
        if self._frames_received[telegram_idx] >= block_size:
            rx_id = self.can_rx_id(telegram_idx)
            tx_id = self.can_tx_id(telegram_idx)
            min_separation_time = 0 # ms
            fc_payload = bitstruct.pack("u4u4u8u8",
                                        IsoTp.FRAME_TYPE_FLOW_CONTROL,
                                        IsoTp.FLOW_CONTROL_CONTINUE,
                                        block_size,
                                        min_separation_time)
            self._send_can_message(tx_id, fc_payload)

            self._frames_received[telegram_idx] = 0 # TODO: 1?

        super().on_consecutive_frame(telegram_idx, segment_idx, frame_payload)


    def _send_can_message(self, id, payload):
        if len(payload) < self._padding_size:
            payload = bytes(payload)+bytes([self._padding_value]*(self._padding_size - len(payload)))

        msg = can.Message(arbitration_id=id,
                          data=payload,
                          is_extended_id=False)

        self._can_bus.send(msg)



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH


from ..diagcodedtypes import read_diag_coded_type_from_odx
from ..globals import xsi
from ..utils import read_element_id

from .codedconstparameter import CodedConstParameter
from .dynamicparameter import DynamicParameter
from .lengthkeyparameter import LengthKeyParameter
from .matchingrequestparameter import MatchingRequestParameter
from .nrcconstparameter import NrcConstParameter
from .physicalconstantparameter import PhysicalConstantParameter
from .reservedparameter import ReservedParameter
from .systemparameter import SystemParameter
from .tableentryparameter import TableEntryParameter
from .tablekeyparameter import TableKeyParameter
from .tablestructparameter import TableStructParameter
from .valueparameter import ValueParameter


def read_parameter_from_odx(et_element):
    element_id = read_element_id(et_element)
    short_name = element_id["short_name"]
    long_name = element_id.get("long_name")
    description = element_id.get("description")

    semantic = et_element.get("SEMANTIC")

    byte_position = int(et_element.find(
        "BYTE-POSITION").text) if et_element.find("BYTE-POSITION") is not None else None
    bit_position = int(et_element.find(
        "BIT-POSITION").text) if et_element.find("BIT-POSITION") is not None else 0

    parameter_type = et_element.get(f"{xsi}type")

    # Which attributes are set depends on the type of the parameter.
    if parameter_type in ["VALUE", "PHYS-CONST", "SYSTEM", "LENGTH-KEY"]:
        dop_ref = et_element.find(
            "DOP-REF").get("ID-REF") if et_element.find("DOP-REF") is not None else None
        dop_snref = et_element.find(
            "DOP-SNREF").get("SHORT-NAME") if et_element.find("DOP-SNREF") is not None else None

        if dop_ref is None and dop_snref is None:
            raise ValueError(
                f"A parameter of type {parameter_type} must reference a DOP! {dop_ref}, {dop_snref}")

    if parameter_type == "VALUE":
        physical_default_value = et_element.find("PHYSICAL-DEFAULT-VALUE").text \
            if et_element.find("PHYSICAL-DEFAULT-VALUE") is not None else None

        return ValueParameter(short_name=short_name,
                              long_name=long_name,
                              semantic=semantic,
                              byte_position=byte_position,
                              bit_position=bit_position,
                              dop_ref=dop_ref,
                              dop_snref=dop_snref,
                              physical_default_value=physical_default_value,
                              description=description)

    elif parameter_type == "PHYS-CONST":
        physical_constant_value = et_element.find(
            "PHYS-CONSTANT-VALUE").text

        return PhysicalConstantParameter(short_name,
                                         long_name=long_name,
                                         semantic=semantic,
                                         byte_position=byte_position,
                                         bit_position=bit_position,
                                         dop_ref=dop_ref,
                                         dop_snref=dop_snref,
                                         physical_constant_value=physical_constant_value,
                                         description=description)

    elif parameter_type == "CODED-CONST":
        diag_coded_type = read_diag_coded_type_from_odx(
            et_element.find("DIAG-CODED-TYPE"))
        coded_value = diag_coded_type.base_data_type.from_string(
            et_element.find("CODED-VALUE").text)

        return CodedConstParameter(short_name,
                                   long_name=long_name,
                                   semantic=semantic,
                                   diag_coded_type=diag_coded_type,
                                   coded_value=coded_value,
                                   byte_position=byte_position,
                                   bit_position=bit_position,
                                   description=description)

    elif parameter_type == "NRC-CONST":
        diag_coded_type = read_diag_coded_type_from_odx(
            et_element.find("DIAG-CODED-TYPE"))
        coded_values = [diag_coded_type.base_data_type.from_string(val.text)
                        for val in et_element.iterfind("CODED-VALUES/CODED-VALUE")]

        return NrcConstParameter(short_name,
                                 long_name=long_name,
                                 semantic=semantic,
                                 diag_coded_type=diag_coded_type,
                                 coded_values=coded_values,
                                 byte_position=byte_position,
                                 bit_position=bit_position,
                                 description=description)

    elif parameter_type == "RESERVED":
        bit_length = int(et_element.find("BIT-LENGTH").text)

        return ReservedParameter(short_name,
                                 long_name=long_name,
                                 semantic=semantic,
                                 byte_position=byte_position,
                                 bit_position=bit_position,
                                 bit_length=bit_length,
                                 description=description)

    elif parameter_type == "MATCHING-REQUEST-PARAM":
        byte_length = int(et_element.find("BYTE-LENGTH").text)
        request_byte_pos = int(
            et_element.find("REQUEST-BYTE-POS").text)

        return MatchingRequestParameter(short_name,
                                        long_name=long_name,
                                        semantic=semantic,
                                        byte_position=byte_position, bit_position=bit_position,
                                        request_byte_position=request_byte_pos, byte_length=byte_length,
                                        description=description)

    elif parameter_type == "SYSTEM":
        sysparam = et_element.get("SYSPARAM")

        return SystemParameter(short_name=short_name,
                               sysparam=sysparam,
                               long_name=long_name,
                               semantic=semantic,
                               byte_position=byte_position,
                               bit_position=bit_position,
                               dop_ref=dop_ref,
                               dop_snref=dop_snref,
                               description=description)

    elif parameter_type == "LENGTH-KEY":
        id = et_element.get("ID")

        return LengthKeyParameter(short_name=short_name,
                                  id=id,
                                  long_name=long_name,
                                  semantic=semantic,
                                  byte_position=byte_position,
                                  bit_position=bit_position,
                                  dop_ref=dop_ref,
                                  dop_snref=dop_snref,
                                  description=description)

    elif parameter_type == "DYNAMIC":

        return DynamicParameter(short_name=short_name,
                                long_name=long_name,
                                semantic=semantic,
                                byte_position=byte_position,
                                bit_position=bit_position,
                                description=description)

    elif parameter_type == "TABLE-STRUCT":
        key_ref = et_element.find(
            "TABLE-KEY-REF").get("ID-REF") if et_element.find("TABLE-KEY-REF") is not None else None
        key_snref = et_element.find(
            "TABLE-KEY-SNREF").get("SHORT-NAME") if et_element.find("TABLE-KEY-SNREF") is not None else None

        return TableStructParameter(short_name=short_name,
                                    table_key_ref=key_ref,
                                    table_key_snref=key_snref,
                                    long_name=long_name,
                                    semantic=semantic,
                                    byte_position=byte_position,
                                    bit_position=bit_position,
                                    description=description)

    elif parameter_type == "TABLE-KEY":

        table_ref = et_element.find(
            "TABLE-REF").get("ID-REF") if et_element.find("TABLE-REF") is not None else None
        table_snref = et_element.find(
            "TABLE-SNREF").get("SHORT-NAME") if et_element.find("TABLE-SNREF") is not None else None
        row_snref = et_element.find(
            "TABLE-ROW-SNREF").get("SHORT-NAME") if et_element.find("TABLE-ROW-SNREF") is not None else None
        row_ref = et_element.find(
            "TABLE-ROW-REF").get("ID-REF") if et_element.find("TABLE-ROW-REF") is not None else None

        return TableKeyParameter(short_name=short_name,
                                 table_ref=table_ref,
                                 table_snref=table_snref,
                                 table_row_snref=row_snref,
                                 table_row_ref=row_ref,
                                 long_name=long_name,
                                 semantic=semantic,
                                 byte_position=byte_position,
                                 bit_position=bit_position,
                                 description=description)

    elif parameter_type == "TABLE-ENTRY":
        target = et_element.find("TARGET").text
        table_row_ref = et_element.find("TABLE-ROW-REF").get("ID-REF")

        return TableEntryParameter(short_name=short_name,
                                   target=target,
                                   table_row_ref=table_row_ref,
                                   long_name=long_name,
                                   byte_position=byte_position,
                                   bit_position=bit_position,
                                   semantic=semantic,
                                   description=description)

    raise NotImplementedError(f"I don't know the type {parameter_type}")



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

import sys
import argparse
from typing import Dict, List, Union
import PyInquirer

from ..database import Database
from ..service import DiagService
from ..structures import Request, Response
from ..parameters import ParameterWithDOP
from ..odxtypes import DataType
from . import _parser_utils

import logging
# logging.basicConfig(level=logging.DEBUG)


def _convert_string_to_odx_type(string_value: str, odx_type: DataType):
    """Similar to odx_type.from_string(string_value) but more relaxed to parse user input"""
    if odx_type == DataType.A_UINT32:
        return int(string_value, 0)
    elif odx_type == DataType.A_BYTEFIELD:
        return _convert_string_to_bytes(string_value)
    else:
        return odx_type.from_string(string_value)


def _convert_string_to_bytes(string_value):
    if all(len(x) <= 2 for x in string_value.split(" ")):
        return bytes(int(x, 16) for x in string_value.split(" ") if len(x) > 0)
    else:
        return int(string_value, 16).to_bytes((int(string_value, 16).bit_length() + 7) // 8, 'big')


def _validate_string_value(input, parameter):
    if parameter.is_optional() and input == "":
        return True
    elif isinstance(parameter, ParameterWithDOP):
        try:
            val = _convert_string_to_odx_type(input,
                                              parameter.physical_type.base_data_type
                                              )
        except:
            return False
        return parameter.dop.is_valid_physical_value(val)
    else:
        logging.info(
            "This value is not validated precisely: Parameter {parameter}"
        )
        return input != ""


def prompt_single_parameter_value(parameter):
    if not isinstance(parameter, ParameterWithDOP) or parameter.get_valid_physical_values() is None or parameter.get_valid_physical_values() is []:
        param_prompt = [{
            "type": "input",
            "name": parameter.short_name,
            "message": f"Value for parameter '{parameter.short_name}' (Type: {parameter.physical_type.base_data_type})" + (f"[optional]" if parameter.is_optional() else ""),
            # TODO: improve validation
            "validate": lambda x: _validate_string_value(x, parameter),
            # TODO: do type conversion?
            "filter": lambda x: x
            # x if x == "" or p.physical_type.base_data_type is None
            # else _convert_string_to_odx_type(x, p.physical_type.base_data_type, param=p) # This does not work because the next parameter to be promted is used (for some reason?)
        }]
    else:
        param_prompt = [{
            "type": "list",
            "name": parameter.short_name,
            "message": f"Value for parameter '{parameter.short_name}'",
            "choices": parameter.get_valid_physical_values()
        }]
    answer = PyInquirer.prompt(param_prompt)
    if answer.get(parameter.short_name) == "" and parameter.is_optional():
        return None
    elif parameter.physical_type.base_data_type is not None:
        return _convert_string_to_odx_type(answer.get(parameter.short_name), parameter.physical_type.base_data_type)
    else:
        logging.warning(
            f"Parameter {parameter.short_name} does not have a physical data type. Param details: {parameter}")
        return answer.get(parameter.short_name)


def encode_message_interactively(sub_service, ask_user_confirmation=False):
    if not sys.__stdin__.isatty() or not sys.__stdout__.isatty():
        raise SystemError(
            "This command can only be used in an interactive shell!")
    param_dict = sub_service.parameter_dict()

    # list(filter(lambda p: p.is_required()
    #                          or p.is_optional(), sub_service.parameters))
    exists_definable_param = False
    for k, param_or_dict in param_dict.items():
        if isinstance(param_or_dict, dict):
            for k, param in param_or_dict.items():
                if param.is_required() or param.is_optional():
                    exists_definable_param = True
        elif param_or_dict.is_required() or param_or_dict.is_optional():
            exists_definable_param = True

    param_values = {}
    if exists_definable_param > 0:
        # Ask whether user wants to encode a message
        if ask_user_confirmation:
            encode_message_prompt = [{
                "type": "list",
                "name": "yes_no_prompt",
                "message": f"Do you want to encode a message? [y/n]",
                "choices": ["yes", "no"]
            }]
            answer = PyInquirer.prompt(encode_message_prompt)
            if answer.get("yes_no_prompt") == "no":
                return

        if isinstance(sub_service, Response):
            answered_request_prompt = [{
                "type": "input",
                "name": "request",
                "message": f"What is the request you want to answer? (Enter the coded request as integer in hexadecimal format (e.g. 12 3B 5)",
                "filter": lambda input: _convert_string_to_bytes(input)
            }]
            answer = PyInquirer.prompt(answered_request_prompt)
            answered_request = answer.get("request")
            print(f"Input interpretation as list: {list(answered_request)}")

        # Request values for parameters
        for key, param_or_structure in param_dict.items():
            if isinstance(param_or_structure, dict):
                # param_or_structure refers to a structure (represented as dict of params)
                print(
                    f"The next {len(param_or_structure)} parameters belong to the structure '{key}'")
                structure_param_values = {}
                for param_sn, param in param_or_structure.items():
                    if param.is_required() or param.is_optional():
                        val = prompt_single_parameter_value(param)
                        if val is not None:
                            structure_param_values[param_sn] = val
                param_values[key] = structure_param_values
            elif (param_or_structure.is_required() or param_or_structure.is_optional()) and param_or_structure.parameter_type != "MATCHING-REQUEST-PARAM":
                # param_or_structure is a parameter
                val = prompt_single_parameter_value(param_or_structure)
                if val is not None:
                    param_values[key] = val
        if isinstance(sub_service, Response):
            payload = sub_service.encode(coded_request=answered_request,
                                         **param_values)
        else:
            payload = sub_service.encode(**param_values)
    else:
        # There are no optional parameters that need to be defined by the user -> Just print message
        payload = sub_service.encode()
    print(f"Message payload: 0x{bytes(payload).hex()}")


def encode_message_from_string_values(sub_service: Union[Request, Response],
                                      parameter_values: Dict[str, Union[str, Dict[str, str]]] = {}):
    parameter_values = parameter_values.copy()
    param_dict = sub_service.parameter_dict()

    # Check if all needed parameters are given
    missing_parameter_names = []
    for parameter_sn, parameter in param_dict.items():
        if isinstance(parameter, dict):
            # parameter_value refers to a structure (represented as dict of params)
            for simple_param_sn, simple_param in parameter.items():
                structured_value = parameter_values.get(parameter_sn)
                if simple_param.is_required() and (not isinstance(structured_value, dict)
                                                   or structured_value.get(simple_param_sn) is None):
                    missing_parameter_names.append(
                        f"{parameter_sn} :: {simple_param_sn}")
        else:
            if parameter.is_required() and parameter_values.get(parameter_sn) is None:
                missing_parameter_names.append(parameter_sn)
    if len(missing_parameter_names) > 0:
        print("The following parameters are required but missing!")
        print(" - " + "\n - ".join(missing_parameter_names))
        return

    # Request values for parameters
    for parameter_sn, parameter_value in parameter_values.items():
        if isinstance(parameter_value, dict):
            # parameter_value refers to a structure (represented as dict of params)
            typed_dict = parameter_value.copy()
            for simple_param_sn, simple_val in parameter_value.items():
                try:
                    parameter = param_dict[parameter_sn][simple_param_sn]
                except:
                    print(f"I don't know the parameter {simple_param_sn}")
                    continue
                typed_dict[simple_param_sn] = _convert_string_to_odx_type(
                    simple_val,
                    parameter.physical_type.base_data_type
                )
                parameter_values[parameter_sn] = typed_dict
        else:
            try:
                parameter = param_dict[parameter_sn]
            except:
                print(f"I don't know the parameter {parameter_sn}")
                continue
            parameter_values[parameter_sn] = _convert_string_to_odx_type(
                parameter_value,
                parameter.physical_type.base_data_type if parameter.parameter_type != "MATCHING-REQUEST-PARAM"
                else DataType.A_BYTEFIELD
            )
    payload = sub_service.encode(**parameter_values) # type: ignore
    print(f"Message payload: 0x{bytes(payload).hex()}")


def browse(odxdb: Database):
    if not sys.__stdin__.isatty() or not sys.__stdout__.isatty():
        raise SystemError(
            "This command can only be used in an interactive shell!")
    dl_names = list(map(lambda dl: dl.short_name, odxdb.diag_layers))
    while True:
        # Select an ECU
        selection = [{
            "type": "list",
            "name": "variant",
            "message": "Select a Variant.",
            "choices": list(dl_names) + ["[exit]"]
        }]
        answer = PyInquirer.prompt(selection)
        if answer.get("variant") == "[exit]":
            return

        variant = odxdb.diag_layers[answer.get("variant")]
        assert variant is not None
        recv_id = hex(variant.get_receive_id()
                      ) if variant.get_receive_id() is not None else "None"
        send_id = hex(variant.get_send_id()
                      ) if variant.get_send_id() is not None else "None"
        print(
            f"{variant.variant_type} '{variant.short_name}' (Receive ID: {recv_id}, Send ID: {send_id})"
        )

        service_sn = 0
        while True:
            services: List[DiagService] = [s for s in variant.services if isinstance(s, DiagService)]
            # Select a service of the ECU
            selection = [{
                "type": "list",
                "name": "service",
                "message": f"The variant {variant.short_name} offers the following services. Select one!",
                "choices": [s.short_name for s in services] + ["[back]"],
            }]
            answer = PyInquirer.prompt(selection)
            if answer.get("service") == "[back]":
                break

            service_sn = answer.get("service")

            service = variant.services[service_sn]
            assert service is not None and isinstance(service, DiagService)
            assert service.request is not None
            assert service.positive_responses is not None
            assert service.negative_responses is not None

            # Select a request/ response of the service
            selection = [{
                "type": "list",
                "name": "message_type",
                "message": "This service offers the following messages.",
                "choices":
                    [{
                        "name": f"Request: {service.request.short_name}",
                        "value": service.request,
                        "short": f"Request: {service.request.short_name}"
                    }] +
                    [{
                        "name": f"Positive response: {pr.short_name}",
                        "value": pr,
                        "short": f"Positive response: {pr.short_name}"
                    } for pr in service.positive_responses] +
                    [{
                        "name": f"Negative response: {nr.short_name}",
                        "value": nr,
                        "short": f"Negative response: {nr.short_name}"
                    } for nr in service.negative_responses]
                    + ["[back]"] # type: ignore
            }]
            answer = PyInquirer.prompt(selection)
            if answer.get("message_type") == "[back]":
                continue

            sub_service = answer.get("message_type")
            sub_service.print_message_format()

            encode_message_interactively(
                sub_service, ask_user_confirmation=True)


def add_subparser(subparsers):
    # Browse interactively to avoid spamming the console.
    parser = subparsers.add_parser(
        'browse',
        description="Interactively browse the content of automotive diagnostic files (*.pdx).",
        help="Interactively browse the content of automotive diagnostic files.",
        formatter_class=argparse.RawTextHelpFormatter)

    _parser_utils.add_pdx_argument(parser)


def run(args):
    odxdb = _parser_utils.load_file(args)
    browse(odxdb)



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

"""odxtools is a set of utilities to deal with automotive diagnostic
descriptions of electronic control units using ODX-formatted
description files.

ODX stands for "Open Diagnostic Data Exchange" `[ODX]`_ and is an XML
based file format to describe the diagnostic capabilities of the
various electronic control units (ECUs) of complex distributed
technical systems (mainly cars and trucks). ODX is an open standard,
maintained by ASAM e.V. `[ASAM]`_ and standardized internationally by
ISO 22901 `[ISO22901]`_.

The services provided by odxtools encompass parsing and internalizing
ODX diagnostic database files as well as decoding and endcoding raw
data of diagnostic requests and responses that are send to/received
from ECUs in an pythonic manner.

Examples:
=========

- Load a database from file ``my_diag_db.pdx``:

.. code-block:: python

    import odxtools

    db = odxtools.load_file("my_diag_db.pdx")

- List the names of all available services of the first ECU:

.. code-block:: python

    # [...]

    dl = db.diag_layer_containers[0].diag_layers[0]
    print(f"Available services for {dl.short_name}: {[s.short_name for s in dl.services]}")

- Determine the CAN IDs which the first described ECU uses to send and receive diagnostic messages:

.. code-block:: python

    # [...]

    print(f"ECU listens on ID 0x{dl.get_receive_id():x}")
    print(f"ECU transmits on ID 0x{dl.get_send_id():x}")

- Decode a diagnostic message for an ECU:

.. code-block:: python

    # [...]

    raw_data = b"\xB0\xA7"
    decoded_message = dl.decode(raw_data)
    print(f"decoded message: {decoded_message}")

References
==========

- _`[ASAM]` ASAM e.V: https://www.asam.net/
- _`[ODX]` The ODX Specification: https://www.asam.net/standards/detail/mcd-2-d/
- _`[ISO22901]` The ISO 22901 Standard: https://www.iso.org/standard/41207.html

"""
from . import database
from .database import Database

from .diaglayer import DiagLayer

from .service import DiagService
from .singleecujob import SingleEcuJob

from .structures import Request, Response

from .parameters import CodedConstParameter
from .parameters import PhysicalConstantParameter
from .parameters import ReservedParameter
from .parameters import ValueParameter
from .parameters import MatchingRequestParameter
from .parameters import SystemParameter
from .parameters import LengthKeyParameter
from .parameters import DynamicParameter
from .parameters import TableStructParameter
from .parameters import TableKeyParameter
from .parameters import TableEntryParameter

from .compumethods import IdenticalCompuMethod
from .compumethods import TexttableCompuMethod
from .compumethods import LinearCompuMethod
from .compumethods import ScaleLinearCompuMethod

from .write_pdx_file import write_pdx_file
from .load_pdx_file import load_pdx_file
from .load_odx_d_file import load_odx_d_file
from .load_file import load_file

from .version import __version__
from .exceptions import *

__author__ = 'Katrin Bauer'

def _main():
    # Command line tool
    from .cli import main as _main
    _main.start_cli()



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from enum import Enum
from typing import Any, Callable, Dict, Literal, Type, Union


def bytefield_to_bytearray(bytefield: str) -> bytearray:
    bytes_string = [bytefield[i:i+2] for i in range(0, len(bytefield), 2)]
    return bytearray(map(lambda x: int(x, 16), bytes_string))

PythonType = Union[str, int, float, bytearray]
LiteralPythonType = Type[Union[str, int, float, bytearray]]

_ODX_TYPE_PARSER: Dict[str, Callable[[str], PythonType]] = {
    "A_INT32": int,
    "A_UINT32": int,
    "A_FLOAT32": float,
    "A_FLOAT64": float,
    "A_UNICODE2STRING": str,
    "A_BYTEFIELD": bytefield_to_bytearray,
    # only in DATA-TYPE not in PHYSICAL-DATA-TYPE
    "A_ASCIISTRING": str,
    "A_UTF8STRING": str
}

_ODX_TYPE_TO_PYTHON_TYPE: Dict[str, LiteralPythonType] = {
    "A_INT32": int,
    "A_UINT32": int,
    "A_FLOAT32": float,
    "A_FLOAT64": float,
    "A_UNICODE2STRING": str,
    "A_BYTEFIELD": bytearray,
    # only in DATA-TYPE not in PHYSICAL-DATA-TYPE
    "A_ASCIISTRING": str,
    "A_UTF8STRING": str
}


class DataType(Enum):
    """Types for the physical and internal value.

    These types can be used either as BASE-DATA-TYPE (for the "internal value")
    or the PHYSICAL-DATA-TYPE (for the "physical value").

    Relevant pages in the ASAM MCD-2D specification:
    * p. 38 (Table 1): ASAM types correspondence with XML Schema types
    * p. 96: Restrictions to the bit length given the BASE-DATA-TYPE.
    """
    A_INT32 = "A_INT32"
    A_UINT32 = "A_UINT32"
    A_FLOAT32 = "A_FLOAT32"
    A_FLOAT64 = "A_FLOAT64"
    A_UNICODE2STRING = "A_UNICODE2STRING"
    A_BYTEFIELD = "A_BYTEFIELD"
    # only in DATA-TYPE not in PHYSICAL-DATA-TYPE
    A_ASCIISTRING = "A_ASCIISTRING"
    A_UTF8STRING = "A_UTF8STRING"

    def as_python_type(self) -> type:
        return _ODX_TYPE_TO_PYTHON_TYPE[self.value]

    def from_string(self, value: str) -> Union[int, float, str, bytearray]:
        return _ODX_TYPE_PARSER[self.value](value)

    def make_from(self, value: Any) -> Union[int, float, str, bytearray]:
        if isinstance(value, str):
            # parse the string
            return self.from_string(value)
        else:
            # regular type cast of python objects
            return self.as_python_type()(value)

    def isinstance(self, value: Any) -> bool:
        expected_type = self.as_python_type()
        if isinstance(value, expected_type):
            return True
        elif expected_type == float and isinstance(value, (int, float)):
            return True
        elif self == DataType.A_BYTEFIELD and isinstance(value, (bytearray, bytes)):
            return True
        else:
            return False



#! /usr/bin/python3
#
# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

import sys
import argparse
import can
import asyncio

from . import _parser_utils
import odxtools
import odxtools.uds as uds
import odxtools.isotp_state_machine as ism

odx_diag_layer = None
last_request = None

ecu_rx_id = None
ecu_tx_id = None


def handle_telegram(telegram_id, payload):
    global odx_diag_layer
    global last_request

    if telegram_id == ecu_tx_id:
        if uds.is_reponse_pending(payload):
            print(f" -> response pending")
            return

        decoded_message = None
        if last_request is not None:
            try:
                decoded_message = odx_diag_layer.decode_response(
                    payload, last_request)[0]
            except odxtools.DecodeError:
                pass

        if decoded_message is not None:
            print(f" -> {decoded_message}")
        else:
            print(f" -> unrecognized response of {len(payload)} bytes length: "
                  f"0x{payload.hex()}")

        return

    decoded_message = None
    if odx_diag_layer is not None:
        try:
            decoded_message = odx_diag_layer.decode(payload)[0]
            last_request = payload
        except odxtools.DecodeError:
            last_request = None

    if decoded_message:
        print(f"Tester: {decoded_message}")
    else:
        print(f"Tester: "
              f"{payload.hex()} "
              f"({payload}, {len(payload)} bytes)")


def init_verbose_state_machine(BaseClass, *args, **kwargs):
    class InformativeIsoTpDecoder(BaseClass):
        def on_sequence_error(self, telegram_idx, expected_idx, rx_idx):
            rx_can_id = self.can_rx_id(telegram_idx)
            print(f"Sequence error for ID 0x{rx_can_id:x}: "
                  f"Received sequence number {rx_idx} but expected {expected_idx}")

            super(BaseClass, self).on_sequence_error(
                telegram_idx, expected_idx, rx_idx)

        def on_frame_type_error(self, telegram_idx, frame_type):
            rx_can_id = self.can_rx_id(telegram_idx)

            print(
                f"Invalid ISO-TP frame for CAN ID 0x{rx_can_id:x}: {frame_type}")

    return InformativeIsoTpDecoder(*args, **kwargs)


async def active_main(args):
    global ecu_rx_id, ecu_tx_id

    can_bus = can.Bus(channel=args.channel, bustype="socketcan")

    ecu_rx_id = int(args.rx, 0)
    ecu_tx_id = int(args.tx, 0)
    isotp_decoder = init_verbose_state_machine(BaseClass=ism.IsoTpActiveDecoder,
                                               can_bus=can_bus,
                                               can_rx_ids=ecu_rx_id,
                                               can_tx_ids=ecu_tx_id,
                                               padding_size=8)

    print(f"Reacting to messages on channel {args.channel}")
    async for telegram_id, payload in isotp_decoder.read_telegrams(can_bus):
        handle_telegram(telegram_id, payload)


async def passive_main(args):
    global ecu_rx_id, ecu_tx_id

    ecu_rx_id = int(args.rx, 0)
    ecu_tx_id = int(args.tx, 0)
    isotp_decoder = init_verbose_state_machine(BaseClass=ism.IsoTpStateMachine,
                                               can_rx_ids=[ecu_rx_id, ecu_tx_id])

    if args.channel:
        # decode a "real" bus
        can_bus = can.Bus(channel=args.channel, bustype="socketcan")

        print(f"Decoding messages on channel {args.channel}")
        async for telegram_id, payload in isotp_decoder.read_telegrams(can_bus):
            handle_telegram(telegram_id, payload)
    else:
        # decode data from stdin
        async for telegram_id, payload in isotp_decoder.read_telegrams(sys.stdin):
            handle_telegram(telegram_id, payload)


def add_cli_arguments(parser):
    parser.add_argument("--active", "-a", metavar="active_mode", default=False, action='store_const', const=True, required=False,
                        help="Active mode, sends flow control messages to receive ISO-TP telegrams successfully")
    parser.add_argument("--channel", "-c", default=None,
                        help="CAN interface name to be used (required in active mode)")
    parser.add_argument("--rx", "-r", default=None, required=False,
                        help="CAN ID in which the ECU listens for diagnostic messages")
    parser.add_argument("--tx", "-t", default=None, required=False,
                        help="CAN ID in which the ECU sends replys to diagnostic messages  (required in active mode)")
    parser.add_argument("--variant", "-v", default=None, required=False,
                        help="Name of the ECU variant which the decode process ought to be based on")
    _parser_utils.add_pdx_argument(parser)


def add_subparser(subparsers):
    parser = subparsers.add_parser(
        'snoop',
        description="Live decoding of a diagnostic session.",
        help="Live decoding of a diagnostic session.",
        formatter_class=argparse.RawTextHelpFormatter)
    add_cli_arguments(parser)


def run(args, odx_database=None):
    global odx_diag_layer
    odx_database = _parser_utils.load_file(args)

    if (odx_database is not None or args.variant is not None) and \
       (odx_database is None or args.variant is None):
        print("The database and variant must always either be "
              "both specified or not at all")
        sys.exit(1)

    odx_diag_layer = None
    if odx_database is not None:
        odx_diag_layer = odx_database.diag_layers[args.variant]

        if odx_diag_layer is None:
            print(
                f"Variant '{args.variant}' does not exist. Available variants:")
            for dl in odx_database.diag_layers:
                print(f"  {dl.short_name}")
            sys.exit(1)

    # if no can IDs have been explicitly specified, take them from the DL
    if args.rx is None and odx_diag_layer:
        args.rx = str(odx_diag_layer.get_receive_id())

    if args.tx is None and odx_diag_layer:
        args.tx = str(odx_diag_layer.get_send_id())

    if args.rx is None:
        print(
            f"If no database and variant is specified, a CAN receive ID must be provided.")
        sys.exit(1)

    if args.active:
        asyncio.run(active_main(args))
    else:
        asyncio.run(passive_main(args))


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Decode ISO-TP communication over a can bus")

    add_cli_arguments(parser)

    args = parser.parse_args()  # deals with the help message handling

    odx_database = odxtools.load_pdx_file(args.pdx_file)

    run(args, odx_database)



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from .nameditemlist import NamedItemList
from .companydata import CompanyData, TeamMember
from .utils import read_description_from_odx

from dataclasses import dataclass, field
from typing import Optional, List

@dataclass()
class CompanyDocInfo:
    company_data_ref: str
    company_data: Optional[CompanyData] = None
    team_member_ref: Optional[str] = None
    team_member: Optional[TeamMember] = None
    doc_label: Optional[str] = None

    def _resolve_references(self, id_lookup):
        self.company_data = id_lookup[self.company_data_ref]

        if self.team_member_ref is not None:
            self.team_member = id_lookup[self.team_member_ref]

@dataclass()
class Modification:
    change: Optional[str] = None
    reason: Optional[str] = None

@dataclass()
class DocRevision:
    """
    Representation of a single revision of the relevant object.
    """
    date: str
    team_member_ref: Optional[str] = None
    team_member: Optional[TeamMember] = None
    revision_label: Optional[str] = None
    state: Optional[str] = None
    tool: Optional[str] = None
    modifications: List[Modification] = field(default_factory=list)

    def _resolve_references(self, id_lookup):
        if self.team_member_ref is not None:
            self.team_member = id_lookup[self.team_member_ref]

@dataclass()
class AdminData:
    language: Optional[str] = None
    company_doc_infos: Optional[List[CompanyDocInfo]] = None
    doc_revisions: Optional[List[DocRevision]] = None

    def _build_id_lookup(self):
        result = {}

        return result

    def _resolve_references(self, id_lookup):
        if self.company_doc_infos is not None:
            for cdi in self.company_doc_infos:
                cdi._resolve_references(id_lookup)

        if self.doc_revisions is not None:
            for dr in self.doc_revisions:
                dr._resolve_references(id_lookup)

def read_admin_data_from_odx(et_element):
    if et_element is None:
        return None

    language = et_element.find("LANGUAGE")
    if language is not None:
        language = language.text

    company_doc_infos = et_element.find("COMPANY-DOC-INFOS")
    if company_doc_infos is not None:
        cdilist = list()
        for cdi in company_doc_infos.iterfind("COMPANY-DOC-INFO"):
            # the company data reference is mandatory
            company_data_ref = cdi.find("COMPANY-DATA-REF").attrib["ID-REF"]

            team_member_ref = cdi.find("TEAM-MEMBER-REF")
            if team_member_ref is not None:
                team_member_ref = team_member_ref.attrib["ID-REF"]

            doc_label = cdi.find("DOC-LABEL")
            if doc_label is not None:
                doc_label = doc_label.text

            cdilist.append(CompanyDocInfo(company_data_ref=company_data_ref,
                                          team_member_ref=team_member_ref,
                                          doc_label=doc_label))

        company_doc_infos = cdilist

    doc_revisions = et_element.find("DOC-REVISIONS")
    if doc_revisions is not None:
        drlist = list()
        for dr in doc_revisions.iterfind("DOC-REVISION"):
            team_member_ref = dr.find("TEAM-MEMBER-REF")
            if team_member_ref is not None:
                team_member_ref = team_member_ref.attrib["ID-REF"]

            revision_label = dr.find("REVISION-LABEL")
            if revision_label is not None:
                revision_label = revision_label.text

            state = dr.find("STATE")
            if state is not None:
                state = state.text

            date = dr.find("DATE").text

            tool = dr.find("TOOL")
            if tool is not None:
                tool = tool.text

            modlist = None
            mods = dr.find("MODIFICATIONS")
            if mods is not None:
                modlist = list()
                for mod in mods.iterfind("MODIFICATION"):
                    m_change = mod.find("CHANGE")
                    if m_change is not None:
                        m_change = m_change.text

                    m_reason = mod.find("REASON")
                    if m_reason is not None:
                        m_reason = m_reason.text

                    modlist.append(Modification(change=m_change,
                                                reason=m_reason))

            drlist.append(DocRevision(team_member_ref=team_member_ref,
                                      revision_label=revision_label,
                                      state=state,
                                      date=date,
                                      tool=tool,
                                      modifications=modlist))


        doc_revisions = drlist

    else:
        doc_revisions = []

    return AdminData(language=language,
                     company_doc_infos=company_doc_infos,
                     doc_revisions=doc_revisions)



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from dataclasses import dataclass, field
from typing import List, Literal, Optional, Union

from .nameditemlist import NamedItemList
from .utils import read_description_from_odx

UnitGroupCategory = Literal["COUNTRY", "EQUIV-UNITS"]


@dataclass
class PhysicalDimension:
    """A physical dimension is a formal definition of a unit.

    It consists of the exponents for the SI units:

    | Symbol | Name     | Quantity                  | Property                 |
    |  ---   | ---      | ---                       | ---                      |
    | s      | second   | time                      | `time_exp`               |
    | m      | metre    | length                    | `length_exp`             |
    | kg     | kilogram | mass                      | `mass_exp`               |
    | A      | ampere   | electric current          | `current_exp`            |
    | K      | kelvin   | thermodynamic temperature | `temperature_exp`        |
    | mol    | mole     | amount of substance       | `molar_amount_exp`       |
    | cd     | candela  | luminous intensity        | `luminous_intensity_exp` |

    (The first three columns are from https://en.wikipedia.org/wiki/International_System_of_Units.)


    Examples
    --------

    The unit `m/s` (or `m**1 * s**(-1)`) can be represented as
    ```
    PhysicalDimension(
        id="velocity",
        short_name="metre_per_second",
        length_exp=1,
        time_exp=-1
    )
    ```
    """
    id: str
    short_name: str
    oid: Optional[str] = None
    long_name: Optional[str] = None
    description: Optional[str] = None
    length_exp: int = 0
    mass_exp: int = 0
    time_exp: int = 0
    current_exp: int = 0
    temperature_exp: int = 0
    molar_amount_exp: int = 0
    luminous_intensity_exp: int = 0


@dataclass
class Unit:
    """
    A unit consists of an ID, short name and a display name.

    Additionally, a unit may reference an SI unit (`.physical_dimension`)
    and an offset to that unit (`factor_si_to_unit`, `offset_si_to_unit`).
    The factor and offset are defined such that the following equation holds true:

    UNIT = FACTOR-SI-TO-UNIT * SI-UNIT + OFFSET-SI-TO-UNIT

    For example: 1km = 1000 * 1m + 0

    Examples
    --------

    A minimal unit representing kilometres:

    ```
    Unit(
        id="kilometre",
        short_name="kilometre",
        display_name="km"
    )
    ```

    A unit that also references a physical dimension:

    ```
    Unit(
        id="ID.kilometre",
        short_name="Kilometre",
        display_name="km",
        physical_dimension_ref="ID.metre",
        factor_si_to_unit=1000,
        offset_si_to_unit=0
    )
    # where the physical_dimension_ref references, e.g.:
    PhysicalDimension(id="ID.metre", short_name="metre", length_exp=1)
    ```
    """
    id: str
    short_name: str
    display_name: str
    oid: Optional[str] = None
    long_name: Optional[str] = None
    description: Optional[str] = None
    factor_si_to_unit: Optional[float] = None
    offset_si_to_unit: Optional[float] = None
    physical_dimension_ref: Optional[str] = None

    def __post_init__(self):
        self._physical_dimension = None

        if self.factor_si_to_unit is not None or self.offset_si_to_unit is not None or self.physical_dimension_ref is not None:
            assert self.factor_si_to_unit is not None and self.offset_si_to_unit is not None and self.physical_dimension_ref is not None, (
                f"Error 54: If one of factor_si_to_unit, offset_si_to_unit and physical_dimension_ref is defined,"
                f" all of them must be defined: {self.factor_si_to_unit} and {self.offset_si_to_unit} and {self.physical_dimension_ref}"
            )

    @property
    def physical_dimension(self) -> PhysicalDimension:
        return self._physical_dimension

    def _resolve_references(self, id_lookup):
        if self.physical_dimension_ref:
            self._physical_dimension = id_lookup[self.physical_dimension_ref]

            assert isinstance(self._physical_dimension, PhysicalDimension), (
                f"The physical_dimension_ref must be resolved to a PhysicalDimension."
                f" {self.physical_dimension_ref} referenced {self._physical_dimension}"
            )


@dataclass
class UnitGroup:
    """A group of units.

    There are two categories of groups: COUNTRY and EQUIV-UNITS.
    """
    short_name: str
    category: UnitGroupCategory
    unit_refs: List[str] = field(default_factory=list)
    oid: Optional[str] = None
    long_name: Optional[str] = None
    description: Optional[str] = None

    def __post_init__(self):
        self._units = NamedItemList[Unit](lambda unit: unit.short_name)

    def _resolve_references(self, id_lookup):
        self._units = NamedItemList[Unit](
            lambda unit: unit.short_name,
            [id_lookup[ref] for ref in self.unit_refs]
        )

    @property
    def units(self) -> NamedItemList[Unit]:
        return self._units


@dataclass
class UnitSpec:
    """
    A unit spec encapsulates three lists:

    * unit groups
    * units
    * physical_dimensions

    The following odx elements are not internalized: ADMIN-DATA, SDGS
    """
    # TODO (?): Why are there type errors...
    unit_groups: Union[NamedItemList[UnitGroup],
                       List[UnitGroup]] = field(default_factory=list)  # type: ignore
    units: Union[NamedItemList[Unit], List[Unit]] = field(
        default_factory=list)  # type: ignore
    physical_dimensions: Union[NamedItemList[PhysicalDimension],
                               List[PhysicalDimension]] = field(default_factory=list)  # type: ignore

    def __post_init__(self):
        self.unit_groups = NamedItemList(lambda x: x.short_name,
                                         self.unit_groups)
        self.units = NamedItemList(lambda x: x.short_name, self.units)
        self.physical_dimensions = NamedItemList(lambda x: x.short_name,
                                                 self.physical_dimensions)

    def _build_id_lookup(self):
        id_lookup = {}
        id_lookup.update({
            unit.id: unit for unit in self.units
        })
        id_lookup.update({
            dim.id: dim for dim in self.physical_dimensions
        })
        return id_lookup

    def _resolve_references(self, id_lookup):
        for unit in self.units:
            unit._resolve_references(id_lookup)
        for group in self.unit_groups:
            group._resolve_references(id_lookup)


def read_unit_from_odx(et_element):
    id = et_element.get("ID")
    oid = et_element.get("OID")
    short_name = et_element.find("SHORT-NAME").text
    long_name = et_element.findtext("LONG-NAME")
    description = read_description_from_odx(et_element.find("DESC"))
    display_name = et_element.find("DISPLAY-NAME").text

    def read_optional_float(element, name):
        if element.findtext(name):
            return float(element.findtext(name))
        else:
            return None
    factor_si_to_unit = read_optional_float(et_element, "FACTOR-SI-TO-UNIT")
    offset_si_to_unit = read_optional_float(et_element, "OFFSET-SI-TO-UNIT")

    ref_element = et_element.find("PHYSICAL-DIMENSION-REF")
    if ref_element is not None:
        physical_dimension_ref = ref_element.get("ID-REF")
    else:
        physical_dimension_ref = None

    return Unit(
        id=id,
        short_name=short_name,
        display_name=display_name,
        oid=oid,
        long_name=long_name,
        description=description,
        factor_si_to_unit=factor_si_to_unit,
        offset_si_to_unit=offset_si_to_unit,
        physical_dimension_ref=physical_dimension_ref
    )


def read_physical_dimension_from_odx(et_element):
    id = et_element.get("ID")
    oid = et_element.get("OID")
    short_name = et_element.find("SHORT-NAME").text
    long_name = et_element.findtext("LONG-NAME")
    description = read_description_from_odx(et_element.find("DESC"))

    def read_optional_int(element, name):
        if element.findtext(name):
            return int(element.findtext(name))
        else:
            return 0

    length_exp = read_optional_int(et_element, "LENGTH-EXP")
    mass_exp = read_optional_int(et_element, "MASS-EXP")
    time_exp = read_optional_int(et_element, "TIME-EXP")
    current_exp = read_optional_int(et_element, "CURRENT-EXP")
    temperature_exp = read_optional_int(et_element, "TEMPERATURE-EXP")
    molar_amount_exp = read_optional_int(et_element, "MOLAR-AMOUNT-EXP")
    luminous_intensity_exp = read_optional_int(et_element,
                                               "LUMINOUS-INTENSITY-EXP")

    return PhysicalDimension(
        id=id,
        short_name=short_name,
        oid=oid,
        long_name=long_name,
        description=description,
        length_exp=length_exp,
        mass_exp=mass_exp,
        time_exp=time_exp,
        current_exp=current_exp,
        temperature_exp=temperature_exp,
        molar_amount_exp=molar_amount_exp,
        luminous_intensity_exp=luminous_intensity_exp
    )


def read_unit_group_from_odx(et_element):
    oid = et_element.get("OID")
    short_name = et_element.find("SHORT-NAME").text
    long_name = et_element.findtext("LONG-NAME")
    description = read_description_from_odx(et_element.find("DESC"))
    category = et_element.findtext("CATEGORY")
    assert category in [
        "COUNTRY", "EQUIV-UNITS"], f'A UNIT-GROUP-CATEGORY must be "COUNTRY" or "EQUIV-UNITS". It was {category}.'
    unit_refs = [el.get("ID-REF")
                 for el in et_element.iterfind("UNIT-REFS/UNIT-REF")]

    return UnitGroup(
        short_name=short_name,
        category=category,
        unit_refs=unit_refs,
        oid=oid,
        long_name=long_name,
        description=description
    )


def read_unit_spec_from_odx(et_element):

    unit_groups = [read_unit_group_from_odx(el)
                   for el in et_element.iterfind("UNIT-GROUPS/UNIT-GROUP")]
    units = [read_unit_from_odx(el)
             for el in et_element.iterfind("UNITS/UNIT")]
    physical_dimensions = [read_physical_dimension_from_odx(el)
                           for el in et_element.iterfind("PHYSICAL-DIMENSIONS/PHYSICAL-DIMENSION")]
    return UnitSpec(
        unit_groups=unit_groups,
        units=units,
        physical_dimensions=physical_dimensions
    )



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH


import abc
from typing import Union

from ..odxtypes import DataType


class CompuMethod:

    def __init__(self,
                 internal_type: Union[DataType, str],
                 physical_type: Union[DataType, str],
                 category: str):
        self.internal_type = DataType(internal_type)
        self.physical_type = DataType(physical_type)
        self.category = category

    @abc.abstractclassmethod
    def convert_physical_to_internal(self, physical_value):
        raise NotImplementedError()

    @abc.abstractclassmethod
    def convert_internal_to_physical(self, internal_value):
        raise NotImplementedError()

    @abc.abstractclassmethod
    def is_valid_physical_value(self, physical_value):
        raise NotImplementedError()

    @abc.abstractclassmethod
    def is_valid_internal_value(self, internal_value):
        raise NotImplementedError()

    def get_valid_physical_values(self):
        return None



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH


from typing import Iterable

from ..globals import logger

from .compumethodbase import CompuMethod
from .linearcompumethod import LinearCompuMethod


class ScaleLinearCompuMethod(CompuMethod):
    def __init__(self, linear_methods: Iterable[LinearCompuMethod]):
        super().__init__(list(linear_methods)[0].internal_type,
                         list(linear_methods)[0].physical_type,
                         "SCALE-LINEAR")
        self.linear_methods = list(linear_methods)
        logger.debug("Created scale linear compu method!")

    def convert_physical_to_internal(self, physical_value):
        assert self.is_valid_physical_value(physical_value), \
            f"cannot convert the invalid physical value {physical_value} of type {type(physical_value)}"
        lin_method = next(
            scale for scale in self.linear_methods if scale.is_valid_physical_value(physical_value))
        return lin_method.convert_physical_to_internal(physical_value)

    def convert_internal_to_physical(self, internal_value):
        lin_method = next(
            scale for scale in self.linear_methods if scale.is_valid_internal_value(internal_value))
        return lin_method.convert_internal_to_physical(internal_value)

    def is_valid_physical_value(self, physical_value):
        return any(True for scale in self.linear_methods if scale.is_valid_physical_value(physical_value))

    def is_valid_internal_value(self, internal_value):
        return any(True for scale in self.linear_methods if scale.is_valid_internal_value(internal_value))



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH


from ..decodestate import DecodeState
from ..exceptions import DecodeError

from .parameterbase import Parameter


class ReservedParameter(Parameter):
    def __init__(self,
                 short_name,
                 bit_length,
                 **kwargs):
        super().__init__(short_name,
                         parameter_type="RESERVED",
                         **kwargs)
        self._bit_length = bit_length

    @property
    def bit_length(self):
        return self._bit_length

    def is_required(self):
        return False

    def is_optional(self):
        return False

    def get_coded_value(self):
        return 0

    def get_coded_value_as_bytes(self, encode_state):
        return int(0).to_bytes((self.bit_length + self.bit_position + 7) // 8, "big")

    def decode_from_pdu(self, decode_state: DecodeState):
        byte_position = self.byte_position if self.byte_position is not None else decode_state.next_byte_position
        byte_length = (self.bit_length + self.bit_position + 7) // 8
        val_as_bytes = decode_state.coded_message[byte_position:byte_position+byte_length]
        next_byte_position = byte_position + byte_length

        # Check that reserved bits are 0
        expected = sum(2**i for i in range(self.bit_position,
                       self.bit_position + self.bit_length))
        actual = int.from_bytes(val_as_bytes, "big")

        # Bit-wise compare if reserved bits are 0.
        if expected & actual != 0:
            raise DecodeError(
                f"Reserved bits must be Zero! "
                f"The parameter {self.short_name} expected {self.bit_length} bits to be Zero starting at bit position {self.bit_position} "
                f"at byte position {byte_position} "
                f"in coded message {decode_state.coded_message.hex()}."
            )

        return None, next_byte_position

    def __repr__(self):
        repr_str = f"ReservedParameter(short_name='{self.short_name}'"
        if self.long_name is not None:
            repr_str += f", long_name='{self.long_name}'"
        if self.byte_position is not None:
            repr_str += f", byte_position='{self.byte_position}'"
        if self.bit_position:
            repr_str += f", bit_position='{self.bit_position}'"
        if self.bit_length is not None:
            repr_str += f", bit_length='{self.bit_length}'"
        if self.semantic is not None:
            repr_str += f", semantic='{self.semantic}'"
        if self.description is not None:
            repr_str += f", description='{' '.join(self.description.split())}'"
        return repr_str + ")"



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from typing import Callable, Dict, Iterable, List, Optional, Union, Generic, TypeVar

T = TypeVar('T')


class NamedItemList(Generic[T]):
    """A list that provides direct access to its items as named attributes.

    This is a hybrid between a list and a user-defined object: One can
    iterate over all items of the list as usual, but items can also be
    accessed via `named_list.itemname`, where the 'itemname' is
    specified via a item -> string mapping function that is passed to
    the constructor.

    If an item name is not unique, `_<num>` will be appended to
    avoid naming collisions. The user is responsible that the strings
    returned by the item-to-name function are valid identifiers in python.
    """

    def __init__(self, item_to_name_fn: Callable[[T], str], input_list: Iterable[T] = None):
        self._item_to_name_fn = item_to_name_fn
        self._list: List[T] = []
        # TODO (?): This duplicates self.__dict__ -> Is there a prettier type-safe way?
        self._typed_dict: Dict[str, T] = {}

        if input_list is not None:
            for item in input_list:
                self.append(item)

    def append(self, item: T):
        """
        Append a new item to the list and make it accessible as a
        member attribute.

        \return The name under which item is accessible
        """
        self._list.append(item)

        item_name = self._item_to_name_fn(item)
        i = 1
        tmp = item_name
        while True:
            if tmp not in self.__dict__:
                self.__dict__[tmp] = item
                self._typed_dict[tmp] = item
                return tmp

            i += 1
            tmp = f"{item_name}_{i}"

    def sort(self, key=None, reverse=False):
        return self._list.sort(key=key, reverse=reverse)

    def __len__(self):
        return len(self._list)

    def __getitem__(self, key: Union[int, str]) -> Optional[T]:
        if isinstance(key, int):
            return self._list[key]
        else:
            return self._typed_dict.get(key)

    def __eq__(self, other: object) -> bool:
        """
        Named item lists are equal if the underlying lists are equal.
        Note that this does not consider the map `item_to_name_fn`.
        """
        if not isinstance(other, NamedItemList):
            return False
        else:
            return self._list == other._list

    def __iter__(self):
        return iter(self._list)

    def __str__(self):
        return f"[{', '.join([self._item_to_name_fn(s) for s in self._list])}]"

    def __repr__(self):
        return self.__str__()



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

import abc
from typing import List, Optional, Union
from dataclasses import dataclass

from .physicaltype import PhysicalType, read_physical_type_from_odx
from .globals import logger
from .compumethods import CompuMethod, read_compu_method_from_odx
from .units import Unit
from .diagcodedtypes import DiagCodedType, StandardLengthType, read_diag_coded_type_from_odx
from .decodestate import DecodeState
from .encodestate import EncodeState
from .exceptions import DecodeError, EncodeError


class DopBase(abc.ABC):
    """ Base class for all DOPs.

    Any class that a parameter can reference via a DOP-REF should inherit from this class.
    """

    def __init__(self,
                 id,
                 short_name,
                 long_name=None,
                 description=None,
                 is_visible=True):
        self.id = id
        self.short_name = short_name
        self.long_name = long_name
        self.description = description
        self.is_visible = is_visible

    @abc.abstractclassmethod
    def convert_physical_to_bytes(self, physical_value, encode_state: EncodeState, bit_position: int) -> bytes:
        """Convert the physical value into bytes."""
        pass

    @abc.abstractclassmethod
    def convert_bytes_to_physical(self, decode_state: DecodeState, bit_position: int = 0):
        """Extract the bytes from the PDU and convert them to the physical value."""
        pass


class DataObjectProperty(DopBase):
    """This class represents a DATA-OBJECT-PROP."""

    def __init__(self,
                 id: str,
                 short_name: str,
                 diag_coded_type: DiagCodedType,
                 physical_type: PhysicalType,
                 compu_method: CompuMethod,
                 unit_ref: Optional[str] = None,
                 long_name: Optional[str] = None,
                 description: Optional[str] = None
                 ):
        super().__init__(id=id,
                         short_name=short_name,
                         long_name=long_name,
                         description=description)
        self.diag_coded_type = diag_coded_type
        self.physical_type = physical_type
        self.compu_method = compu_method
        self.unit_ref = unit_ref
        self._unit = None

    @property
    def unit(self) -> Optional[Unit]:
        return self._unit

    @property
    def bit_length(self):
        # TODO: The DiagCodedTypes except StandardLengthType don't have a bit length.
        #       Should we remove this bit_length property from DOP or return None?
        if isinstance(self.diag_coded_type, StandardLengthType):
            return self.diag_coded_type.bit_length
        else:
            return None

    def convert_physical_to_internal(self, physical_value):
        assert self.physical_type.base_data_type.isinstance(physical_value),\
            f"Expected {self.physical_type.base_data_type.value}, got {type(physical_value)}"

        return self.compu_method.convert_physical_to_internal(physical_value)

    def convert_physical_to_bytes(self, physical_value, encode_state, bit_position):
        if not self.is_valid_physical_value(physical_value):
            raise EncodeError(f"The value {repr(physical_value)} of type {type(physical_value)}"
                              f" is not a valid." +
                              (f" Valid values are {self.compu_method.get_valid_physical_values()}"
                               if self.compu_method.get_valid_physical_values()
                               else f" Expected type {self.physical_type.base_data_type.value}.")
                              )

        internal_val = self.convert_physical_to_internal(physical_value)
        return self.diag_coded_type.convert_internal_to_bytes(internal_val, encode_state, bit_position=bit_position)

    def convert_bytes_to_physical(self, decode_state, bit_position: int = 0):
        assert 0 <= bit_position and bit_position < 8

        internal, next_byte_position = self.diag_coded_type.convert_bytes_to_internal(decode_state,
                                                                                      bit_position=bit_position)

        if self.compu_method.is_valid_internal_value(internal):
            return self.compu_method.convert_internal_to_physical(internal), next_byte_position
        else:
            # TODO: How to prevent this?
            raise DecodeError(f"DOP {self.short_name} could not convert the coded value "
                              f" {repr(internal)} to physical type {self.physical_type.base_data_type}.")

    def is_valid_physical_value(self, physical_value):
        return self.compu_method.is_valid_physical_value(physical_value)

    def get_valid_physical_values(self):
        return self.compu_method.get_valid_physical_values()

    def _resolve_references(self, id_lookup):
        """Resolves the reference to the unit"""
        if self.unit_ref:
            self._unit = id_lookup[self.unit_ref]

    def __repr__(self) -> str:
        return \
            f"DataObjectProperty('{self.short_name}', " + \
            ", ".join([
                f"category='{self.compu_method.category}'",
                f"internal_type='{self.diag_coded_type}'",
                f"physical_type={self.physical_type}",
                f"unit_ref='{self.unit_ref}'",
            ]) + \
            ")"

    def __str__(self) -> str:
        return \
            f"DataObjectProperty('{self.short_name}', " + \
            ", ".join([
                f"category='{self.compu_method.category}'",
                f"diag_coded_type='{self.diag_coded_type}'",
                f"physical_type={self.physical_type}",
                f"unit_ref='{self.unit_ref}'",
            ]) + \
            ")"


@dataclass
class DiagnosticTroubleCode:
    trouble_code: int
    id: Optional[str] = None
    short_name: Optional[str] = None
    text: Optional[str] = None
    display_trouble_code: Optional[str] = None
    level: Union[bytes, bytearray, None] = None
    is_temporary: bool = False


class DtcRef:
    """A proxy for DiagnosticTroubleCode.
    The DTC is referenced by ID and the ID-REF
    is resolved after loading the pdx database.
    """

    def __init__(self, dtc_id):
        self.dtc_id = dtc_id
        self.dtc: Optional[DiagnosticTroubleCode] = None

    @property
    def id(self):
        return self.dtc.id

    @property
    def short_name(self):
        return self.dtc.short_name

    @property
    def trouble_code(self):
        return self.dtc.trouble_code

    @property
    def text(self):
        return self.dtc.text

    @property
    def display_trouble_code(self):
        return self.dtc.display_trouble_code

    @property
    def level(self):
        return self.dtc.level

    @property
    def is_temporary(self):
        return self.dtc.is_temporary

    def _resolve_references(self, id_lookup):
        self.dtc: Optional[DiagnosticTroubleCode] = id_lookup.get(self.dtc_id)
        if self.dtc is None:
            logger.debug(f"DTC-REF {self.dtc_id} could not be resolved.")
        else:
            assert isinstance(self.dtc, DiagnosticTroubleCode),\
                f"DTC-REF {self.dtc_id} does not reference a DTC but a {type(self.dtc)}."


class DtcDop(DataObjectProperty):
    """ A DOP describing a diagnostic trouble code """

    def __init__(self,
                 id: str,
                 short_name: str,
                 diag_coded_type: DiagCodedType,
                 physical_type: PhysicalType,
                 compu_method: CompuMethod,
                 dtcs: List[Union[DiagnosticTroubleCode, DtcRef]],
                 is_visible=False,
                 linked_dtc_dops=False,
                 long_name: str = None,
                 description: str = None
                 ):
        super().__init__(id=id,
                         short_name=short_name,
                         long_name=long_name,
                         description=description,
                         diag_coded_type=diag_coded_type,
                         physical_type=physical_type,
                         compu_method=compu_method)
        self.dtcs = dtcs
        self.is_visible = is_visible
        self.linked_dtc_dops = linked_dtc_dops

    def convert_bytes_to_physical(self, decode_state, bit_position: int = 0):
        trouble_code, next_byte = super().convert_bytes_to_physical(decode_state,
                                                                    bit_position=bit_position)

        dtcs = [ x for x in self.dtcs if x.trouble_code == trouble_code ]

        assert len(dtcs) < 2, \
            f"Multiple matching DTCs for trouble code 0x{trouble_code:06x}"

        if len(dtcs) == 1:
            # we found exactly one described DTC
            return dtcs[0], next_byte

        # the DTC was not specified. This probably means that the
        # diagnostic description file is incomplete. We do not bail
        # out but we cannot provide an interpretation for it out of the
        # box...
        dtc = DiagnosticTroubleCode(trouble_code=trouble_code)

        return dtc, next_byte

    def convert_physical_to_bytes(self, physical_value, encode_state, bit_position):
        if isinstance(physical_value, DiagnosticTroubleCode):
            trouble_code = physical_value.trouble_code
        elif isinstance(physical_value, int):
            # assume that physical value is the trouble_code
            trouble_code = physical_value
        elif isinstance(physical_value, str):
            # assume that physical value is the short_name
            dtc = next(filter(lambda dtc: dtc.short_name == physical_value,
                              self.dtcs))
            trouble_code = dtc.trouble_code
        else:
            raise EncodeError(f"The DTC-DOP {self.short_name} expected a"
                              f" DiagnosticTroubleCode but got {physical_value}.")

        return super().convert_physical_to_bytes(trouble_code, encode_state, bit_position)

    def _build_id_lookup(self):
        id_lookup = {}
        id_lookup[self.id] = self
        for dtc in self.dtcs:
            if isinstance(dtc, DiagnosticTroubleCode):
                id_lookup[dtc.id] = dtc
        return id_lookup

    def _resolve_references(self, id_lookup):
        for dtc in self.dtcs:
            if isinstance(dtc, DtcRef):
                dtc._resolve_references(id_lookup)


def read_dtc_from_odx(et_element):
    if et_element.find("DISPLAY-TROUBLE-CODE") is not None:
        display_trouble_code = et_element.find("DISPLAY-TROUBLE-CODE").text
    else:
        display_trouble_code = None

    if et_element.find("LEVEL") is not None:
        level = et_element.find("LEVEL").text
    else:
        level = None

    return DiagnosticTroubleCode(id=et_element.get("ID"),
                                 short_name=et_element.find("SHORT-NAME").text,
                                 trouble_code=int(
                                     et_element.find("TROUBLE-CODE").text),
                                 text=et_element.find("TEXT").text,
                                 display_trouble_code=display_trouble_code,
                                 level=level)


def read_data_object_property_from_odx(et_element):
    """Reads a DATA-OBJECT-PROP or a DTC-DOP."""
    id = et_element.get("ID")
    short_name = et_element.find("SHORT-NAME").text
    long_name = et_element.find("LONG-NAME").text
    description = et_element.find("DESCRIPTION").text if et_element.find(
        "DESCRIPTION") is not None else None
    logger.debug('Parsing DOP ' + short_name)

    diag_coded_type = read_diag_coded_type_from_odx(
        et_element.find("DIAG-CODED-TYPE"))

    physical_type = read_physical_type_from_odx(
        et_element.find("PHYSICAL-TYPE"))
    compu_method = read_compu_method_from_odx(et_element.find(
        "COMPU-METHOD"), diag_coded_type.base_data_type, physical_type.base_data_type)

    if et_element.tag == "DATA-OBJECT-PROP":
        if et_element.find("UNIT-REF") is not None:
            unit_ref = et_element.find("UNIT-REF").get("ID-REF")
        else:
            unit_ref = None
        dop = DataObjectProperty(id,
                                 short_name,
                                 diag_coded_type,
                                 physical_type,
                                 compu_method,
                                 unit_ref=unit_ref,
                                 long_name=long_name,
                                 description=description)
    else:
        dtcs = [read_dtc_from_odx(el)
                for el in et_element.iterfind("DTCS/DTC")]
        dtcs += [DtcRef(el.get("ID-REF"))
                 for el in et_element.iterfind("DTCS/DTC-REF")]

        is_visible = et_element.get("IS-VISIBLE") == "true"
        dop = DtcDop(id,
                     short_name,
                     diag_coded_type,
                     physical_type,
                     compu_method,
                     dtcs,
                     is_visible=is_visible,
                     long_name=long_name,
                     description=description)
    return dop



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from typing import Dict, Literal
from xml.etree import ElementTree


def read_description_from_odx(et_element: ElementTree.Element):
    """Read a DESCRIPTION element. The element usually has the name DESC."""
    # TODO: Invent a better representation of a DESC element.
    #       This just represents it as XHTML string. 
    if et_element is None:
        return None

    raw_string = ElementTree.tostring(et_element,
                                      encoding="unicode",
                                      method="xml").strip()
    # Remove DESC start and end tag.
    assert raw_string.startswith("<DESC>"), raw_string
    assert raw_string.endswith("</DESC>"), raw_string
    # Remove starting and ending tag
    raw_string = raw_string[len("<DESC>"):-len("</DESC>")]
    return raw_string.strip()


def read_element_id(et_element) -> Dict[Literal["short_name", "long_name", "description"], str]:
    """Read the elements "SHORT-NAME", "LONG-NAME" and "DESCRIPTION".

    Returns the dict
    ```python
    {
        "short_name": et_element.find("SHORT-NAME"),
        "long_name": et_element.find("LONG-NAME"),
        "description": read_description_from_odx(et_element.find("DESC"))
    }
    ```
    If `et_element` does not contain the elements "LONG-NAME" and "DESC",
    the returned dict does not contain the correspondig keys.

    A typical use for this function is reading an odx element
    that contains the group "ELEMENT-ID", e.g.,
    ```python
    def read_type_with_element_id_from_odx(et_element):
        element_id = read_element_id(et_element)
        return TypeWithElementID(**element_id)
    ```

    """
    d: Dict[Literal["short_name", "long_name", "description"], str] = {
        "short_name": et_element.findtext("SHORT-NAME")
    }
    if et_element.find("LONG-NAME") is not None:
        d["long_name"] = et_element.findtext("LONG-NAME")
    if et_element.find("DESC") is not None:
        d["description"] = read_description_from_odx(et_element.find("DESC"))
    return d



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH


from ..encodestate import EncodeState

from .parameterwithdop import ParameterWithDOP


class LengthKeyParameter(ParameterWithDOP):
    """Length Keys specify the bit (!) length of another parameter.

    The other parameter references the length key parameter using a ParamLengthInfoType as .diag_coded_type.

    LengthKeyParameters are decoded the same as ValueParameters.
    The main difference to ValueParameters is that a LengthKeyParameter has an `.id` attribute
    and its DOP must be a simple DOP with PHYSICAL-TYPE/BASE-DATA-TYPE="A_UINT32".
    """

    def __init__(self,
                 short_name,
                 id,
                 dop_ref=None,
                 dop_snref=None,
                 long_name=None,
                 byte_position=None,
                 bit_position=0,
                 semantic=None,
                 description=None):
        super().__init__(short_name,
                         parameter_type="LENGTH-KEY",
                         dop_ref=dop_ref,
                         dop_snref=dop_snref,
                         long_name=long_name,
                         byte_position=byte_position,
                         bit_position=bit_position,
                         semantic=semantic,
                         description=description)
        self.id = id

    def is_required(self):
        raise NotImplementedError(
            "LengthKeyParameter.is_required is not implemented yet.")

    def is_optional(self):
        raise NotImplementedError(
            "LengthKeyParameter.is_optional is not implemented yet.")

    def encode_into_pdu(self, encode_state: EncodeState) -> bytearray:
        physical_value = encode_state.parameter_values.get(self.short_name,)

        # TODO: Maybe it'd be nice if the length key value would default to the minimal needed length,
        #       instead of raising a TypeError here. But for that we'd
        #       (1) need to find the parameter that references this length key and determine its bit length
        #       and (2) make sure that we choose a length is a valid physical value for the DOP.
        #       (Restrictions may be e.g. multiple of 8 or 16, depending on the compu method.)
        if physical_value is None:
            raise TypeError(f"A value for the length key '{self.short_name}'"
                            f" must be specified.")

        # Set the value of the length key in the length key dict.
        encode_state.length_keys[self.id] = physical_value

        return super(ParameterWithDOP, self).encode_into_pdu(encode_state)



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from itertools import chain
from typing import Optional, Any, Dict, Iterable, List, Union

from .exceptions import *
from .globals import logger, xsi
from .state import read_state_from_odx
from .state_transition import read_state_transition_from_odx

from .utils import read_description_from_odx
from .nameditemlist import NamedItemList
from .admindata import AdminData, read_admin_data_from_odx
from .companydata import CompanyData, read_company_datas_from_odx
from .communicationparameter import CommunicationParameterRef, read_communication_param_ref_from_odx
from .diagdatadictionaryspec import DiagDataDictionarySpec, read_diag_data_dictionary_spec_from_odx
from .dataobjectproperty import DopBase
from .functionalclass import read_functional_class_from_odx
from .audience import read_additional_audience_from_odx
from .message import Message
from .service import DiagService, read_diag_service_from_odx
from .singleecujob import SingleEcuJob, read_single_ecu_job_from_odx
from .structures import Request, Response, read_structure_from_odx

# Defines priority of overiding objects
PRIORITY_OF_DIAG_LAYER_TYPE = {
    "PROTOCOL": 1,
    "FUNCTIONAL-GROUP": 2,
    "BASE-VARIANT": 3,
    "ECU-VARIANT": 4,
    # Inherited services from ECU Shared Data always override inherited services from other diag layers
    "ECU-SHARED-DATA": 5
}


class DiagLayer:

    class ParentRef:
        def __init__(self,
                     reference,  # : Union[str, DiagLayer],
                     ref_type: str,
                     not_inherited_diag_comms=[],
                     not_inherited_dops=[]):
            """
            Parameters
            ----------
            reference: str | DiagLayer
                the ID (string) or the referenced DiagLayer
            ref_type: str
            not_inherited_diag_comms: List[str]
                short names of not inherited diag comms
            not_inherited_dops: List[str]
                short names of not inherited DOPs
            """
            assert ref_type in ["PROTOCOL-REF", "BASE-VARIANT-REF",
                                "ECU-SHARED-DATA-REF", "FUNCTIONAL-GROUP-REF"]
            if isinstance(reference, str):
                self.id_ref = reference
                self.referenced_diag_layer = None
            else:
                self.id_ref = reference.id
                self.referenced_diag_layer = reference
            self.not_inherited_diag_comms = not_inherited_diag_comms
            self.not_inherited_dops = not_inherited_dops
            self.ref_type = ref_type

        def _resolve_references(self, id_lookup):
            self.referenced_diag_layer = id_lookup.get(self.id_ref)
            if self.referenced_diag_layer is None:
                logger.warning(
                    f"Could not resolve parent ref to {self.id_ref}")

        def get_inheritance_priority(self):
            return PRIORITY_OF_DIAG_LAYER_TYPE[self.referenced_diag_layer.variant_type]

        def get_inherited_services_by_name(self):
            services = {service.short_name: service for service in self.referenced_diag_layer._services
                        if service.short_name not in self.not_inherited_diag_comms}
            return services

        def get_inherited_data_object_properties_by_name(self):
            dops = {dop.short_name: dop for dop in self.referenced_diag_layer._data_object_properties
                    if dop.short_name not in self.not_inherited_dops}
            return dops

        def get_inherited_communication_parameters_by_name(self):
            return {cp.id_ref: cp for cp in self.referenced_diag_layer._communication_parameters}

    def __init__(self,
                 variant_type,
                 id,
                 short_name,
                 long_name=None,
                 description=None,
                 requests: List[Request] = [],
                 positive_responses: List[Response] = [],
                 negative_responses: List[Response] = [],
                 services: List[DiagService] = [],
                 single_ecu_jobs: List[SingleEcuJob] = [],
                 diag_comm_refs: List[str] = [],
                 parent_refs: List[ParentRef] = [],
                 diag_data_dictionary_spec: DiagDataDictionarySpec = None,
                 communication_parameters:
                 List[CommunicationParameterRef] = [],
                 enable_candela_workarounds=True,
                 id_lookup=None,
                 additional_audiences=[],
                 functional_classes=[],
                 states=[],
                 state_transitions=[]
                 ):
        logger.info(f"Initializing variant type {variant_type}")
        self.variant_type = variant_type

        self.id = id
        self.short_name = short_name
        self.long_name = long_name
        self.description = description

        # Requests and Responses
        self.requests = requests
        self.positive_responses = NamedItemList[Response](lambda pr: pr.short_name,
                                                          positive_responses)
        self.negative_responses = NamedItemList[Response](lambda nr: nr.short_name,
                                                          negative_responses)

        # ParentRefs
        self.parent_refs = parent_refs

        # DiagServices (note that they do not include inherited services!)
        self._local_services = NamedItemList[DiagService](lambda ser: ser.short_name,
                                                          services)
        self._local_single_ecu_jobs = NamedItemList[SingleEcuJob](lambda job: job.short_name,
                                                                  single_ecu_jobs)
        self._diag_comm_refs = diag_comm_refs

        # DOP-BASEs
        self.local_diag_data_dictionary_spec = diag_data_dictionary_spec

        # Communication parameters, e.g. CAN-IDs
        self._local_communication_parameters = NamedItemList[CommunicationParameterRef](lambda cp: cp._python_name(),
                                                                                        communication_parameters)

        self.additional_audiences = additional_audiences
        self.functional_classes = functional_classes
        self.states = states
        self.state_transitions = state_transitions

        # Properties that include inherited objects
        self._services: NamedItemList[Union[DiagService, SingleEcuJob]]\
            = NamedItemList(lambda s: s.short_name, [])
        self._communication_parameters: NamedItemList[CommunicationParameterRef]\
            = NamedItemList(lambda s: s.id_ref, [])
        self._data_object_properties: NamedItemList[DopBase]\
            = NamedItemList(lambda s: s.short_name, [])

        if id_lookup is not None:
            self.finalize_init(id_lookup)

        # specify whether enable work arounds for bugs of CANdela studio
        self._enable_candela_workarounds = enable_candela_workarounds

    @property
    def services(self) -> NamedItemList[Union[DiagService, SingleEcuJob]]:
        """All services that this diagnostic layer offers including inherited services."""
        return self._services

    @property
    def data_object_properties(self) -> NamedItemList[DopBase]:
        """All data object properties including inherited ones.
        This attribute corresponds to all specializations of DOP-BASE
        defined in the DIAG-DATA-DICTIONARY-SPEC of this diag layer as well as
        in the DIAG-DATA-DICTIONARY-SPEC of any parent.
        """
        return self._data_object_properties

    @property
    def communication_parameters(self) -> NamedItemList[CommunicationParameterRef]:
        """All communication parameters including inherited ones."""
        return self._communication_parameters

    def finalize_init(self, id_lookup={}):
        """Resolves all references.

        This method should be called whenever the diag layer (or a referenced object) was changed.
        Particularly, this method assumes that all inherited diag layer are correctly initialized,
        i.e., have resolved their references.
        """
        id_lookup.update(self._build_id_lookup())
        self._resolve_references(id_lookup)

    def _build_id_lookup(self):
        """Construct a mapping from IDs to all objects that are contained in this diagnostic layer."""
        logger.info(f"Adding {self.id} to id_lookup.")

        id_lookup = {}
        for obj in chain(self._local_services,
                         self._local_single_ecu_jobs,
                         self.requests,
                         self.positive_responses,
                         self.negative_responses,
                         self.additional_audiences,
                         self.functional_classes,
                         self.states,
                         self.state_transitions):
            id_lookup[obj.id] = obj

        if self.local_diag_data_dictionary_spec:
            id_lookup.update(
                self.local_diag_data_dictionary_spec._build_id_lookup()
            )

        id_lookup[self.id] = self
        return id_lookup

    def _resolve_references(self, id_lookup: Dict[str, Any]) -> None:
        """Recursively resolve all references."""
        # Resolve inheritance
        for pr in self.parent_refs:
            pr._resolve_references(id_lookup)

        services = sorted(self._compute_available_services_by_name(id_lookup).values(),
                          key=lambda service: service.short_name)
        self._services = NamedItemList[Union[DiagService, SingleEcuJob]](
            lambda s: s.short_name,
            services)

        dops = sorted(self._compute_available_data_object_properties_by_name().values(),
                      key=lambda dop: dop.short_name)
        self._data_object_properties = NamedItemList[DopBase](
            lambda dop: dop.short_name,
            dops)

        comparams = sorted(self._compute_available_commmunication_parameters_by_name().values(),
                           key=lambda comparam: comparam.id_ref)
        self._communication_parameters = NamedItemList[CommunicationParameterRef](
            lambda cp: cp._python_name(),
            comparams)

        # Resolve all other references
        for struct in chain(self.requests,
                            self.positive_responses,
                            self.negative_responses):
            struct._resolve_references(self, id_lookup)

        local_diag_comms: Iterable[Union[DiagService, SingleEcuJob]] \
            = (*self._local_services, *self._local_single_ecu_jobs)
        for service in local_diag_comms:
            service._resolve_references(id_lookup)

        if self.local_diag_data_dictionary_spec:
            self.local_diag_data_dictionary_spec._resolve_references(self,
                                                                     id_lookup)

    def __local_services_by_name(self, id_lookup) -> Dict[str, Union[DiagService, SingleEcuJob]]:
        services_by_name: Dict[str, Union[DiagService, SingleEcuJob]] = {}

        for ref in self._diag_comm_refs:
            if ref in id_lookup:
                services_by_name[id_lookup[ref].short_name] = id_lookup[ref]
            else:
                logger.warning(f"Diag comm ref {ref!r} could not be resolved.")

        services_by_name.update({
            service.short_name: service for service in self._local_services
        })
        services_by_name.update({
            service.short_name: service for service in self._local_single_ecu_jobs
        })
        return services_by_name

    def _compute_available_services_by_name(self, id_lookup) -> Dict[str, DiagService]:
        """Helper method for initializing the available services.
        This computes the services that are inherited from other diagnostic layers."""
        services_by_name = {}

        # Look in parent refs for inherited services
        # Fetch services from low priority parents first, then update with increasing priority
        for parent_ref in self._get_parent_refs_sorted_by_priority():
            services_by_name.update(
                parent_ref.get_inherited_services_by_name())

        services_by_name.update(self.__local_services_by_name(id_lookup))
        return services_by_name

    def _compute_available_data_object_properties_by_name(self) -> Dict[str, DopBase]:
        """Returns the locally defined and inherited DOPs."""
        data_object_properties_by_name = {}

        # Look in parent refs for inherited services
        # Fetch services from low priority parents first, then update with increasing priority
        for parent_ref in self._get_parent_refs_sorted_by_priority():
            data_object_properties_by_name.update(
                parent_ref.get_inherited_data_object_properties_by_name())

        if self.local_diag_data_dictionary_spec:
            data_object_properties_by_name.update({
                d.short_name: d for d in self.local_diag_data_dictionary_spec.all_data_object_properties
            })
        return data_object_properties_by_name

    def _compute_available_commmunication_parameters_by_name(self) -> dict:
        com_params_by_name = {}

        # Look in parent refs for inherited services
        # Fetch services from low priority parents first, then update with increasing priority
        for parent_ref in self._get_parent_refs_sorted_by_priority():
            com_params_by_name.update(
                parent_ref.get_inherited_communication_parameters_by_name())

        com_params_by_name.update(
            {cp.id_ref: cp for cp in self._local_communication_parameters})
        return com_params_by_name

    def _get_parent_refs_sorted_by_priority(self, reverse=False):
        return sorted(self.parent_refs, key=lambda pr:  pr.get_inheritance_priority(), reverse=reverse)

    def _build_coded_prefix_tree(self):
        """Constructs the coded prefix tree of the services.
        Each leaf node is a list of `DiagService`s.
        (This is because navigating from a service to the request/ responses is easier than finding the service for a given request/response object.)

        Example:
        Let there be four services with corresponding requests:
        * Request 1 has the coded constant prefix `12 34`.
        * Request 2 has the coded constant prefix `12 34`.
        * Request 3 has the coded constant prefix `12 56`.
        * Request 4 has the coded constant prefix `12 56 00`.

        Then, the constructed prefix tree is the dict
        ```
        {0x12: {0x34: {-1: [<Service 1>, <Service 2>]},
                0x56: {-1: [<Service 3>],
                       0x0: {-1: [<Service 4>]}
                       }}}
        ```
        Note, that the inner `-1` are constant to distinguish them from possible service IDs.

        Also note, that it is actually allowed that
        (a) SIDs for different services are the same like for service 1 and 2 (thus each leaf node is a list) and
        (b) one SID is the prefix of another SID like for service 3 and 4 (thus the constant `-1` key).
        """
        services = list(filter(lambda s: isinstance(
            s, DiagService), self._services))
        prefix_tree = {}
        for s in services:
            # Compute prefixes for the request and all responses
            request_prefix = s.request.coded_const_prefix()
            prefixes = [request_prefix] + [
                message.coded_const_prefix(
                    request_prefix=request_prefix
                ) for message in chain(s.positive_responses, s.negative_responses)
            ]
            for coded_prefix in prefixes:
                # Traverse prefix tree
                sub_tree = prefix_tree
                for b in coded_prefix:
                    if sub_tree.get(b) is None:
                        sub_tree[b] = {}
                    sub_tree = sub_tree.get(b)

                    assert isinstance(
                        sub_tree, dict), f"{sub_tree} has type {type(sub_tree)}. How did this happen?"
                # Add service as leaf to prefix tree
                if sub_tree.get(-1) is None:
                    sub_tree[-1] = [s]
                else:
                    sub_tree[-1].append(s)
        return prefix_tree

    def _find_services_for_uds(self, message: Union[bytes, bytearray]):
        if not hasattr(self, "_prefix_tree"):
            # Compute the prefix tree the first time this decode function is called.
            self._prefix_tree = self._build_coded_prefix_tree()
        prefix_tree = self._prefix_tree

        # Find matching service(s) in prefix tree
        possible_services = []
        for b in message:
            if prefix_tree.get(b) is not None:
                assert isinstance(prefix_tree.get(b), dict)
                prefix_tree = prefix_tree.get(b)
            else:
                break
            if -1 in prefix_tree:
                possible_services += prefix_tree[-1]
        return possible_services

    def decode(self, message: Union[bytes, bytearray]) -> Iterable[Message]:
        possible_services = self._find_services_for_uds(message)

        if possible_services is None:
            raise DecodeError(
                f"Couldn't find corresponding service for message {message.hex()}."
            )

        decoded_messages = []

        for service in possible_services:
            try:
                decoded_messages.append(service.decode_message(message))
            except DecodeError as e:
                pass
        if len(decoded_messages) == 0:
            raise DecodeError(
                f"None of the services {possible_services} could parse {message.hex()}.")
        return decoded_messages

    def decode_response(self, response: Union[bytes, bytearray], request: Union[bytes, bytearray, Message]) -> Iterable[Message]:
        if isinstance(request, Message):
            possible_services = [request.service]
        else:
            if not isinstance(request, (bytes, bytearray)):
                raise TypeError(f"Request parameter must have type "
                                f"Message, bytes or bytearray but was {type(request)}")
            possible_services = self._find_services_for_uds(request)
        if possible_services is None:
            raise DecodeError(
                f"Couldn't find corresponding service for request {request.hex()}."
            )

        decoded_messages = []

        for service in possible_services:
            try:
                decoded_messages.append(service.decode_message(response))
            except DecodeError as e:
                pass
        if len(decoded_messages) == 0:
            raise DecodeError(
                f"None of the services {possible_services} could parse {response.hex()}.")
        return decoded_messages

    def get_receive_id(self):
        """CAN ID to which the ECU listens for diagnostic messages"""
        # TODO (?): When using the dict notation to access comparams,
        #           wouldn't it be prettier to allow the dot in the id?
        com_param = self.communication_parameters["ISO_15765_2__CP_UniqueRespIdTable"]
        if com_param is None:
            return None
        else:
            if self._enable_candela_workarounds:
                # assume the parameter order used by CANdela studio.
                # note that the parameter ordering actually used
                # differs from the one of the COMPARAM fragment
                # delivered by CANdela generated PDX files and that
                # both are different from the one of the COMPARAM
                # fragment included in the MCD2-D standard
                try:
                    return int(com_param.value[2])
                except ValueError:
                    return None

            else:
                # assume the parameter order specified by the COMPARAM
                # fragment of the the ASAM MCD2-D standard.
                return int(com_param.value[1])

    def get_send_id(self):
        """CAN ID to which the ECU sends replies to diagnostic messages"""
        com_param = self.communication_parameters["ISO_15765_2__CP_UniqueRespIdTable"]
        if com_param is None:
            return None
        else:
            if self._enable_candela_workarounds:
                # assume the parameter order used by CANdela studio.
                # note that the parameter odering actually used
                # differs from the one of the COMPARAM fragment
                # delivered by CANdela generated PDX files and that
                # both are different from the one of the COMPARAM
                # fragment included in the MCD2-D standard
                return int(com_param.value[5])
            else:
                # assume the parameter order specified by the COMPARAM
                # fragment of the the ASAM MCD2-D standard.
                return int(com_param.value[4])

    def get_tester_present_time(self):
        """Timeout on inactivity in seconds.

        This is defined by the communication parameter "ISO_14230_3.CP_TesterPresentTime".
        If the variant does not define this parameter, the default value 3.0 is returned.

        Description of the comparam: "Time between a response and the next subsequent tester present message
        (if no other request is sent to this ECU) in case of physically addressed requests."
        """
        cps = list(filter(lambda x: x.id_ref == "ISO_14230_3.CP_TesterPresentTime",
                          self.communication_parameters))
        if len(cps):
            assert len(cps) == 1
            idle_timeout = int(cps[0].value) / 1e6
        else:
            idle_timeout = 3.0  # default specified by the standard
        return idle_timeout

    def __repr__(self) -> str:
        return f"""DiagLayer(variant_type={self.variant_type},
          id={repr(self.id)},
          short_name={repr(self.short_name)},
          long_name={repr(self.long_name)},
          description={repr(self.description)},
          requests={self.requests},
          positive_responses={self.positive_responses},
          negative_responses={self.negative_responses},
          services={self._local_services},
          diag_comm_refs={self._diag_comm_refs},
          parent_refs={self.parent_refs},
          diag_data_dictionary_spec={self.local_diag_data_dictionary_spec},
          communication_parameters={self._local_communication_parameters},
          enable_candela_workarounds={self._enable_candela_workarounds})"""

    def __str__(self) -> str:
        return f"DiagLayer('{self.short_name}', type='{self.variant_type}')"


def read_parent_ref_from_odx(et_element):
    id_ref = et_element.get("ID-REF")

    not_inherited_diag_comms = [el.get("SHORT-NAME")
                                for el in et_element.iterfind("NOT-INHERITED-DIAG-COMMS/NOT-INHERITED-DIAG-COMM/DIAG-COMM-SNREF")]
    not_inherited_dops = [el.get("SHORT-NAME")
                          for el in et_element.iterfind("NOT-INHERITED-DOPS/NOT-INHERITED-DOP/DOP-BASE-SNREF")]
    ref_type = et_element.get(f"{xsi}type")

    return DiagLayer.ParentRef(
        id_ref,
        ref_type=ref_type,
        not_inherited_diag_comms=not_inherited_diag_comms,
        not_inherited_dops=not_inherited_dops
    )


def read_diag_layer_from_odx(et_element, enable_candela_workarounds=True):

    # logger.info(et_element)
    variant_type = et_element.tag

    id = et_element.get("ID")
    short_name = et_element.find("SHORT-NAME").text

    long_name = et_element.find(
        "LONG-NAME").text if et_element.find("LONG-NAME") is not None else None
    description = read_description_from_odx(et_element.find("DESC"))

    logger.info(f"Parsing {variant_type} '{short_name}' ...")

    # Parse DiagServices
    services = [read_diag_service_from_odx(service)
                for service in et_element.iterfind("DIAG-COMMS/DIAG-SERVICE")]
    diag_comm_refs = [service.get("ID-REF")
                      for service in et_element.iterfind("DIAG-COMMS/DIAG-COMM-REF")]
    single_ecu_jobs = [read_single_ecu_job_from_odx(sej)
                       for sej in et_element.iterfind("DIAG-COMMS/SINGLE-ECU-JOB")]

    # Parse ParentRefs
    parent_refs = [read_parent_ref_from_odx(pr_el)
                   for pr_el in et_element.iterfind("PARENT-REFS/PARENT-REF")]

    # Parse communication parameter refs
    com_params = [read_communication_param_ref_from_odx(el)
                  for el in et_element.iterfind("COMPARAM-REFS/COMPARAM-REF")]

    # Parse Requests and Responses
    requests = [read_structure_from_odx(rq)
                for rq in et_element.iterfind("REQUESTS/REQUEST")]
    positive_responses = [read_structure_from_odx(pr)
                          for pr in et_element.iterfind("POS-RESPONSES/POS-RESPONSE")]
    negative_responses = [read_structure_from_odx(nr)
                          for nr in et_element.iterfind("NEG-RESPONSES/NEG-RESPONSE")]

    additional_audiences = [read_additional_audience_from_odx(el)
                            for el in et_element.iterfind("ADDITIONAL-AUDIENCES/ADDITIONAL-AUDIENCE")]

    functional_classes = [
        read_functional_class_from_odx(el) for el in et_element.iterfind("FUNCT-CLASSS/FUNCT-CLASS")]

    states = [
        read_state_from_odx(el) for el in et_element.iterfind("STATE-CHARTS/STATE-CHART/STATES/STATE")]

    state_transitions = [
        read_state_transition_from_odx(el) for el in
        et_element.iterfind("STATE-CHARTS/STATE-CHART/STATE-TRANSITIONS/STATE-TRANSITION")]

    if et_element.find("DIAG-DATA-DICTIONARY-SPEC"):
        diag_data_dictionary_spec = read_diag_data_dictionary_spec_from_odx(
            et_element.find("DIAG-DATA-DICTIONARY-SPEC"))
    else:
        diag_data_dictionary_spec = None

    # TODO: Are UNIT-SPEC and SDGS needed?

    # Create DiagLayer
    dl = DiagLayer(variant_type,
                   id,
                   short_name,
                   long_name=long_name,
                   description=description,
                   requests=requests,
                   positive_responses=positive_responses,
                   negative_responses=negative_responses,
                   services=services,
                   diag_comm_refs=diag_comm_refs,
                   single_ecu_jobs=single_ecu_jobs,
                   parent_refs=parent_refs,
                   diag_data_dictionary_spec=diag_data_dictionary_spec,
                   communication_parameters=com_params,
                   additional_audiences=additional_audiences,
                   functional_classes=functional_classes,
                   states=states,
                   state_transitions=state_transitions,
                   enable_candela_workarounds=enable_candela_workarounds,
                   )

    return dl


class DiagLayerContainer:
    def __init__(self,
                 id,
                 short_name,
                 long_name=None,
                 description=None,
                 admin_data: Optional[AdminData] = None,
                 company_datas: Optional[NamedItemList[CompanyData]] = None,
                 ecu_shared_datas=[],
                 protocols=[],
                 functional_groups=[],
                 base_variants=[],
                 ecu_variants=[]
                 ):
        self.id = id
        self.short_name = short_name
        self.long_name = long_name
        self.description = description
        self.admin_data = admin_data
        self.company_datas = company_datas

        self.ecu_shared_datas = ecu_shared_datas
        self.protocols = protocols
        self.functional_groups = functional_groups
        self.base_variants = base_variants
        self.ecu_variants = ecu_variants

        self._diag_layers = NamedItemList[DiagLayer](
            lambda dop: dop.short_name,
            list(chain(self.ecu_shared_datas, self.protocols, self.functional_groups, self.base_variants, self.ecu_variants)))

    def _build_id_lookup(self):
        result = {}
        if self.admin_data is not None:
            result.update(self.admin_data._build_id_lookup())

        if self.company_datas is not None:
            for cd in self.company_datas:
                result.update(cd._build_id_lookup())

        return result

    def _resolve_references(self, id_lookup: Dict[str, Any]) -> None:
        if self.admin_data is not None:
            self.admin_data._resolve_references(id_lookup)

        if self.company_datas is not None:
            for cd in self.company_datas:
                cd._resolve_references(id_lookup)


    @property
    def diag_layers(self):
        return self._diag_layers

    def __getitem__(self, key: Union[int, str]) -> DiagLayer:
        return self.diag_layers[key]

    def __repr__(self) -> str:
        return f"DiagLayerContainer('{self.short_name}')"

    def __str__(self) -> str:
        return f"DiagLayerContainer('{self.short_name}')"


def read_diag_layer_container_from_odx(et_element, enable_candela_workarounds=True):
    id = et_element.get("ID")
    short_name = et_element.find("SHORT-NAME").text
    try:
        long_name = et_element.find("LONG-NAME").text
    except:
        long_name = None
    description = read_description_from_odx(et_element.find("DESC"))
    admin_data = read_admin_data_from_odx(et_element.find("ADMIN-DATA"))
    company_datas = read_company_datas_from_odx(et_element.find("COMPANY-DATAS"))
    ecu_shared_datas = [read_diag_layer_from_odx(dl_element, enable_candela_workarounds=enable_candela_workarounds)
                        for dl_element in et_element.iterfind("ECU-SHARED-DATAS/ECU-SHARED-DATA")]
    protocols = [read_diag_layer_from_odx(dl_element, enable_candela_workarounds=enable_candela_workarounds)
                 for dl_element in et_element.iterfind("PROTOCOLS/PROTOCOL")]
    functional_groups = [read_diag_layer_from_odx(dl_element, enable_candela_workarounds=enable_candela_workarounds)
                         for dl_element in et_element.iterfind("FUNCTIONAL-GROUPS/FUNCTIONAL-GROUP")]
    base_variants = [read_diag_layer_from_odx(dl_element, enable_candela_workarounds=enable_candela_workarounds)
                     for dl_element in et_element.iterfind("BASE-VARIANTS/BASE-VARIANT")]
    ecu_variants = [read_diag_layer_from_odx(dl_element, enable_candela_workarounds=enable_candela_workarounds)
                    for dl_element in et_element.iterfind("ECU-VARIANTS/ECU-VARIANT")]

    return DiagLayerContainer(id,
                              short_name,
                              long_name=long_name,
                              description=description,
                              admin_data=admin_data,
                              company_datas=company_datas,
                              ecu_shared_datas=ecu_shared_datas,
                              protocols=protocols,
                              functional_groups=functional_groups,
                              base_variants=base_variants,
                              ecu_variants=ecu_variants
                              )



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from dataclasses import dataclass, field
from typing import Any, Dict, List, Literal, Optional, Union, cast

from .dataobjectproperty import DopBase
from .audience import Audience, read_audience_from_odx
from .functionalclass import FunctionalClass
from .utils import read_element_id
from .nameditemlist import NamedItemList
from .globals import logger
from .exceptions import EncodeError, DecodeError
from .message import Message

DiagClassType = Literal["STARTCOMM",
                        "STOPCOMM",
                        "VARIANTIDENTIFICATION",
                        "READ-DYN-DEFINED-MESSAGE",
                        "DYN-DEF-MESSAGE",
                        "CLEAR-DYN-DEF-MESSAGE"]


@dataclass
class InputParam:
    short_name: str
    dop_base_ref: str
    long_name: Optional[str] = None
    description: Optional[str] = None
    oid: Optional[str] = None
    semantic: Optional[str] = None
    physical_default_value: Optional[str] = None

    def __post_init__(self):
        self._dop: Optional[DopBase] = None

    def _resolve_references(self, id_lookup: Dict[str, Any]) -> None:
        self._dop = id_lookup.get(self.dop_base_ref)
        if self._dop is None:
            logger.warning(
                f"DOP-BASE-REF '{self.dop_base_ref!r}' could not be resolved.")

    @property
    def dop(self) -> Optional[DopBase]:
        """The data object property describing this parameter."""
        return self._dop


@dataclass
class OutputParam:
    id: str
    short_name: str
    dop_base_ref: str
    long_name: Optional[str] = None
    description: Optional[str] = None
    oid: Optional[str] = None
    semantic: Optional[str] = None

    def __post_init__(self):
        self._dop: Optional[DopBase] = None

    def _resolve_references(self, id_lookup: Dict[str, Any]) -> None:
        self._dop = id_lookup.get(self.dop_base_ref)
        if self._dop is None:
            logger.warning(
                f"DOP-BASE-REF '{self.dop_base_ref!r}' could not be resolved.")

    @property
    def dop(self) -> Optional[DopBase]:
        """The data object property describing this parameter."""
        return self._dop


@dataclass
class NegOutputParam:
    short_name: str
    dop_base_ref: str
    long_name: Optional[str] = None
    description: Optional[str] = None

    def __post_init__(self):
        self._dop: Optional[DopBase] = None

    def _resolve_references(self, id_lookup: Dict[str, Any]) -> None:
        self._dop = id_lookup.get(self.dop_base_ref)
        if self._dop is None:
            logger.warning(
                f"DOP-BASE-REF '{self.dop_base_ref!r}' could not be resolved.")

    @property
    def dop(self) -> Optional[DopBase]:
        """The data object property describing this parameter."""
        return self._dop


@dataclass
class ProgCode:
    """A reference to code that is executed by a single ECU job
    """
    code_file: str
    syntax: Literal["JAVA", "CLASS", "JAR"]
    revision: str
    encryption: Optional[str] = None
    entrypoint: Optional[str] = None
    library_refs: List[str] = field(default_factory=list)

    def __post_init__(self):
        if not self.library_refs:
            self.library_refs = []

    def _resolve_references(self, id_lookup: Dict[str, Any]) -> None:
        # TODO: Libraries are currently not internalized.
        #       Once they are internalized, resolve the references `library_refs` here.
        pass


@dataclass
class SingleEcuJob:
    """A single ECU job is a diagnostic communication primitive.

    A single ECU job is more complex than a diagnostic service and is not provided natively by the ECU.
    In particular, the job is defined in external programs which are referenced by the attribute `.prog_codes`.

    In contrast to "multiple ECU jobs", a single ECU job only does service calls to a single ECU.

    Single ECU jobs are defined in section 7.3.5.7 of the ASAM MCD-2 standard.

    TODO: The following xml attributes are not internalized yet:
          ADMIN-DATA, SDGS, PROTOCOL-SNREFS, RELATED-DIAG-COMM-REFS, PRE-CONDITION-STATE-REFS, STATE-TRANSITION-REFS
    """
    id: str
    short_name: str
    prog_codes: List[ProgCode]
    """Pointers to the code that is executed when calling this job."""
    # optional xsd:elements inherited from DIAG-COMM (and thus shared with DIAG-SERVICE)
    oid: Optional[str] = None
    long_name: Optional[str] = None
    description: Optional[str] = None
    functional_class_refs: List[str] = field(default_factory=list)
    audience: Optional[Audience] = None
    # optional xsd:elements specific to SINGLE-ECU-JOB
    # cast(...) just tells the type checker to pretend it's a list...
    input_params: Union[NamedItemList[InputParam], List[InputParam]] \
        = cast(List[InputParam], field(default_factory=list))
    output_params: Union[NamedItemList[OutputParam], List[OutputParam]] \
        = cast(List[OutputParam], field(default_factory=list))
    neg_output_params: Union[NamedItemList[NegOutputParam], List[NegOutputParam]] \
        = cast(List[NegOutputParam], field(default_factory=list))
    # xsd:attributes inherited from DIAG-COMM (and thus shared with DIAG-SERVICE)
    semantic: Optional[str] = None
    diagnostic_class: Optional[DiagClassType] = None
    is_mandatory: bool = False
    is_executable: bool = True
    is_final: bool = False

    def __post_init__(self) -> None:
        if not self.functional_class_refs:
            self.functional_class_refs = []
        self._functional_classes: Optional[NamedItemList[FunctionalClass]] = None

        # Replace None attributes by empty lists
        if not self.input_params:
            self.input_params = NamedItemList(lambda x: x.short_name, [])
        if not self.output_params:
            self.output_params = NamedItemList(lambda x: x.short_name, [])
        if not self.neg_output_params:
            self.neg_output_params = NamedItemList(lambda x: x.short_name, [])

    @property
    def functional_classes(self) -> Optional[NamedItemList[FunctionalClass]]:
        """The functional classes referenced by this job.
        This is None iff the references were not resolved.
        """
        return self._functional_classes

    def _resolve_references(self, id_lookup: Dict[str, Any]) -> None:
        # Resolve references to functional classes
        self._functional_classes = NamedItemList[FunctionalClass](
            lambda fc: fc.short_name, [])
        for fc_id in self.functional_class_refs:
            fc = id_lookup.get(fc_id)
            if isinstance(fc, FunctionalClass):
                self._functional_classes.append(fc)
            else:
                logger.warning(
                    f"Functional class ID {fc_id!r} resolved to {fc!r}.")

        # Resolve references of audience
        if self.audience:
            self.audience._resolve_references(id_lookup)

        # Resolve references of params
        params: List[Union[InputParam, OutputParam, NegOutputParam]] \
            = [*self.input_params, *self.output_params, *self.neg_output_params]
        for p in params:
            p._resolve_references(id_lookup)

        for code in self.prog_codes:
            code._resolve_references(id_lookup)

    def decode_message(self, message: Union[bytes, bytearray]) -> Message:
        """This function's signature matches `DiagService.decode_message`
        and only raises an informative error.
        """
        raise DecodeError(f"Single ECU jobs are completely executed on the tester and thus cannot be decoded."
                          f" You tried to decode a response for the job {self.id}.")

    def encode_request(self, **params):
        """This function's signature matches `DiagService.encode_request`
        and only raises an informative error.
        """
        raise EncodeError(f"Single ECU jobs are completely executed on the tester and thus cannot be encoded."
                          f" You tried to encode a request for the job {self.id}.")

    def encode_positive_response(self, coded_request, response_index=0, **params):
        """This function's signature matches `DiagService.encode_positive_response`
        and only raises an informative error.
        """
        raise EncodeError(f"Single ECU jobs are completely executed on the tester and thus cannot be encoded."
                          f" You tried to encode a response for the job {self.id}.")

    def encode_negative_response(self, coded_request, response_index=0, **params):
        """This function's signature matches `DiagService.encode_negative_response`
        and only raises an informative error.
        """
        raise EncodeError(f"Single ECU jobs are completely executed on the tester and thus cannot be encoded."
                          f" You tried to encode the job {self.id}.")

    def __call__(self, **params) -> bytes:
        raise EncodeError(f"Single ECU jobs are completely executed on the tester and thus cannot be encoded."
                          f" You tried to call the job {self.id}.")


def read_prog_code_from_odx(et_element):
    code_file = et_element.find("CODE-FILE").text

    encryption = et_element.findtext("ENCRYPTION")

    syntax = et_element.find("SYNTAX").text
    revision = et_element.find("REVISION").text

    entrypoint = et_element.findtext("ENTRYPOINT")

    library_refs = [
        el.get("ID-REF") for el in et_element.iterfind("LIBRARY-REFS/LIBRARY-REF")
    ]
    return ProgCode(
        code_file=code_file,
        syntax=syntax,
        revision=revision,
        encryption=encryption,
        entrypoint=entrypoint,
        library_refs=library_refs
    )


def read_input_param_from_odx(et_element):
    element_id = read_element_id(et_element)
    dop_base_ref = et_element.find("DOP-BASE-REF").get("ID-REF")
    physical_default_value = et_element.findtext("PHYSICAL-DEFAULT-VALUE")

    # optional attributes
    semantic = et_element.get("SEMANTIC")
    oid = et_element.get("OID")
    return InputParam(
        **element_id,
        dop_base_ref=dop_base_ref,
        physical_default_value=physical_default_value,
        semantic=semantic,
        oid=oid
    )


def read_output_param_from_odx(et_element):
    id = et_element.get("ID")
    element_id = read_element_id(et_element)
    dop_base_ref = et_element.find("DOP-BASE-REF").get("ID-REF")

    # optional attributes
    semantic = et_element.get("SEMANTIC")
    oid = et_element.get("OID")
    return OutputParam(
        id=id,
        **element_id,
        dop_base_ref=dop_base_ref,
        semantic=semantic,
        oid=oid
    )


def read_neg_output_param_from_odx(et_element):
    element_id = read_element_id(et_element)
    dop_base_ref = et_element.find("DOP-BASE-REF").get("ID-REF")

    return NegOutputParam(
        **element_id,
        dop_base_ref=dop_base_ref
    )


def read_single_ecu_job_from_odx(et_element):
    logger.info(
        f"Parsing service based on ET DiagService element: {et_element}")
    id = et_element.attrib["ID"]
    element_id = read_element_id(et_element)
    semantic = et_element.get("SEMANTIC")

    functional_class_ref_ids = [
        el.attrib["ID-REF"] for el in et_element.iterfind("FUNCT-CLASS-REFS/FUNCT-CLASS-REF")
    ]

    prog_codes = [
        read_prog_code_from_odx(el) for el in et_element.iterfind("PROG-CODES/PROG-CODE")
    ]

    if et_element.find("AUDIENCE"):
        audience = read_audience_from_odx(et_element.find("AUDIENCE"))
    else:
        audience = None

    input_params = [
        read_input_param_from_odx(el) for el in et_element.iterfind("INPUT-PARAMS/INPUT-PARAM")
    ]
    output_params = [
        read_output_param_from_odx(el) for el in et_element.iterfind("OUTPUT-PARAMS/OUTPUT-PARAM")
    ]
    neg_output_params = [
        read_neg_output_param_from_odx(el) for el in et_element.iterfind("NEG-OUTPUT-PARAMS/NEG-OUTPUT-PARAM")
    ]

    # Read boolean flags. Note that the "else" clause contains the default value.
    is_mandatory = True if et_element.get("IS-MANDATORY") == "true" else False
    is_executable = (False if et_element.get("IS-EXECUTABLE") == "false"
                     else True)
    is_final = True if et_element.get("IS-FINAL") == "true" else False

    diag_service = SingleEcuJob(id,
                                **element_id,
                                prog_codes=prog_codes,
                                semantic=semantic,
                                audience=audience,
                                functional_class_refs=functional_class_ref_ids,
                                input_params=input_params,
                                output_params=output_params,
                                neg_output_params=neg_output_params,
                                is_mandatory=is_mandatory,
                                is_executable=is_executable,
                                is_final=is_final)
    return diag_service



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

"""Subpackage containing all compu methods.

A computational method is used to convert the physical type
into the internal type and vice versa.

All compu methods inherit from the abstract base class CompuMethod.
Each category is represented by a different sub class.
"""

from .compumethodbase import CompuMethod
from .identicalcompumethod import IdenticalCompuMethod
from .linearcompumethod import LinearCompuMethod
from .scalelinearcompumethod import ScaleLinearCompuMethod
from .tabintpcompumethod import TabIntpCompuMethod
from .texttablecompumethod import TexttableCompuMethod

from .compurationalcoeffs import CompuRationalCoeffs
from .compuscale import CompuScale
from .limit import IntervalType, Limit
from .readcompumethod import read_compu_method_from_odx



#! /usr/bin/python3
#
# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from ..load_file import load_file as _load_file


def add_pdx_argument(parser):
    parser.add_argument(
        "pdx_file", metavar="PDX_FILE", help="path to the .pdx file")
    #parser.add_argument("pdx_files", metavar="PDX_FILES", nargs="+", help="PDX descriptions of all ECUs which shall be analyzed")

    # do not enable the CANdela workarounds. Note that we disable then
    # by default because CANdela is by far the most common tool to
    # work with ODX.
    # Note that the short option "-c" is reserved for the "--channel" of the snoop command.
    parser.add_argument("--conformant", default=False, action='store_const', const=True,
                        required=False, help="The input file fully confirms to the standard, i.e., disable work-arounds for bugs of the CANdela tool")


def load_file(args):
    db_file_name = args.pdx_file
    odxdb = None
    if db_file_name is not None:
        odxdb = _load_file(db_file_name,
                            enable_candela_workarounds=not args.conformant)
    return odxdb



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

class OdxError(Exception):
    """Any error that happens during interacting with diagnostic objects."""

class EncodeError(OdxError):
    """Encoding of a message to raw data failed."""

class DecodeError(OdxError):
    """Decoding raw data failed."""



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from dataclasses import dataclass
from typing import Optional
from odxtools.utils import read_description_from_odx


@dataclass()
class State:
    """
    Corresponds to STATE.
    """
    id: str
    short_name: str
    long_name: Optional[str] = None
    description: Optional[str] = None


def read_state_from_odx(et_element):
    short_name = et_element.find("SHORT-NAME").text
    id = et_element.get("ID")

    long_name = et_element.find(
        "LONG-NAME").text if et_element.find("LONG-NAME") is not None else None
    description = read_description_from_odx(et_element.find("DESC"))

    return State(id=id,
                 short_name=short_name,
                 long_name=long_name,
                 description=description)



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH


from typing import Union

from ..odxtypes import DataType

from .compumethodbase import CompuMethod


class IdenticalCompuMethod(CompuMethod):
    def __init__(self,
                 internal_type: Union[DataType, str],
                 physical_type: Union[DataType, str]):
        super().__init__(internal_type, physical_type, "IDENTICAL")

    def convert_physical_to_internal(self, physical_value):
        return physical_value

    def convert_internal_to_physical(self, internal_value):
        return internal_value

    def is_valid_physical_value(self, physical_value):
        return self.physical_type.isinstance(physical_value)

    def is_valid_internal_value(self, internal_value):
        return self.internal_type.isinstance(internal_value)



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from odxtools.audience import Audience, read_audience_from_odx
from odxtools.functionalclass import FunctionalClass
from odxtools.state import State
from odxtools.utils import read_description_from_odx
from odxtools.exceptions import DecodeError
from typing import List, Optional, Union

from .state_transition import StateTransition
from .structures import Request, Response
from .nameditemlist import NamedItemList
from .message import Message


class DiagService:
    def __init__(self,
                 id,
                 short_name,
                 request,
                 positive_responses: Union[List[str], List[Response]],
                 negative_responses: Union[List[str], List[Response]],
                 long_name=None,
                 description=None,
                 semantic=None,
                 audience: Optional[Audience] = None,
                 functional_class_refs=[],
                 pre_condition_state_refs=[],
                 state_transition_refs=[]):
        """Constructs the service.

        Parameters:
        ----------
        id: str
        short_name: str
            the short name of this DIAG-SERVICE
        request: str | Request
            the ID of a request or a object
        positive_responses: List[str] | List[Response]
        negative_responses: List[str] | List[Response]
        """
        self.id: str = id
        self.short_name: str = short_name
        self.long_name: Optional[str] = long_name
        self.description: Optional[str] = description
        self.semantic: Optional[str] = semantic
        self.audience: Optional[Audience] = audience
        self.functional_class_refs: List[str] = functional_class_refs
        self._functional_classes: Union[List[FunctionalClass],
                                        NamedItemList[FunctionalClass]] = []
        self.pre_condition_state_refs: List[str] = pre_condition_state_refs
        self._pre_condition_states: Union[List[State],
                                          NamedItemList[State]] = []
        self.state_transition_refs: List[str] = state_transition_refs
        self._state_transitions: Union[List[StateTransition],
                                       NamedItemList[StateTransition]] = []

        self._request: Optional[Request]
        self.request_ref_id: str
        self._positive_responses: Optional[NamedItemList[Response]]
        self.pos_res_ref_ids: List[str]
        self._negative_responses: Optional[NamedItemList[Response]]
        self.neg_res_ref_ids: List[str]

        if isinstance(request, str):
            self._request = None
            self.request_ref_id = request
        elif isinstance(request, Request):
            self._request = request
            self.request_ref_id = request.id
        else:
            raise ValueError(
                "request must be a string (the ID of a request) or a Request object")

        if all(isinstance(x, Response) for x in positive_responses):
            # TODO (?): Can we tell mypy that positive_responses is definitely of type Iterable[Response]
            self._positive_responses = \
                NamedItemList[Response](lambda pr: pr.short_name,
                                        positive_responses)  # type: ignore
            self.pos_res_ref_ids = [
                pr.id for pr in positive_responses]  # type: ignore
        elif all(isinstance(x, str) for x in positive_responses):
            self._positive_responses = None
            self.pos_res_ref_ids = [str(r) for r in positive_responses]
        else:
            raise TypeError(
                "positive_responses must be of type Union[List[str], List[Response], None]")

        if all(isinstance(x, Response) for x in negative_responses):
            self._negative_responses = \
                NamedItemList[Response](lambda nr: nr.short_name,
                                        negative_responses)  # type: ignore
            self.neg_res_ref_ids = [
                nr.id for nr in negative_responses]  # type: ignore
        elif all(isinstance(x, str) for x in negative_responses):
            self._negative_responses = None
            self.neg_res_ref_ids = [str(r) for r in negative_responses]
        else:
            raise TypeError(
                "negative_responses must be of type Union[List[str], List[Response], None]")

    @property
    def request(self) -> Optional[Request]:
        return self._request

    @property
    def positive_responses(self) -> Optional[NamedItemList[Response]]:
        return self._positive_responses

    @property
    def negative_responses(self) -> Optional[NamedItemList[Response]]:
        return self._negative_responses

    @property
    def functional_classes(self):
        return self._functional_classes

    @property
    def pre_condition_states(self):
        return self._pre_condition_states

    @property
    def state_transitions(self):
        return self._state_transitions

    def _resolve_references(self, id_lookup):
        self._request = id_lookup.get(self.request_ref_id)
        self._positive_responses = \
            NamedItemList(
                lambda pr: pr.short_name,
                [id_lookup.get(pr_id) for pr_id in self.pos_res_ref_ids])
        self._negative_responses = \
            NamedItemList(
                lambda nr: nr.short_name,
                [id_lookup.get(nr_id) for nr_id in self.neg_res_ref_ids])
        self._functional_classes = \
            NamedItemList(
                lambda fc: fc.short_name,
                [id_lookup.get(fc_id) for fc_id in self.functional_class_refs])
        self._pre_condition_states = \
            NamedItemList(
                lambda st: st.short_name,
                [id_lookup.get(st_id) for st_id in self.pre_condition_state_refs])
        self._state_transitions = \
            NamedItemList(
                lambda st: st.short_name,
                [id_lookup.get(stt_id) for stt_id in self.state_transition_refs])
        if self.audience:
            self.audience._resolve_references(id_lookup)

    def decode_message(self, message: Union[bytes, bytearray]) -> Message:

        # Check if message is a request or positive or negative response
        interpretable_message_types = []

        if self.request is None or self.positive_responses is None or self.negative_responses is None:
            raise ValueError("References couldn't be resolved or have not been resolved yet."
                             " Try calling `database.resolve_references()`.")

        for message_type in [self.request,
                             *self.positive_responses,
                             *self.negative_responses]:
            prefix = message_type.coded_const_prefix(
                request_prefix=self.request.coded_const_prefix())
            if all(b == message[i] for (i, b) in enumerate(prefix)):
                interpretable_message_types.append(message_type)

        if len(interpretable_message_types) != 1:
            raise DecodeError(
                f"The service {self.short_name} cannot decode the message {message.hex()}")
        message_type = interpretable_message_types[0]
        param_dict = message_type.decode(message)
        return Message(coded_message=message, service=self, structure=message_type, param_dict=param_dict)

    def encode_request(self, **params):
        """
        Composes an UDS request as list of bytes for this service.
        Parameters:
        ----------
        params: dict
            Parameters of the RPC as mapping from SHORT-NAME of the parameter to the physical value
        """
        # make sure that all parameters which are required for
        # encoding are specified (parameters which have a default are
        # optional)
        missing_params = set(map(
            lambda x: x.short_name, self.request.get_required_parameters())).difference(params.keys())
        assert not missing_params, f"The parameters {missing_params} are required but missing!"

        # make sure that no unknown parameters are specified
        rq_all_param_names = set(
            map(lambda x: x.short_name, self.request.parameters))
        assert set(params.keys()).issubset(rq_all_param_names), \
            f"Unknown parameters specified for encoding: {params.keys()}, known parameters are: {rq_all_param_names}"
        return self.request.encode(**params)

    def encode_positive_response(self, coded_request, response_index=0, **params):
        # TODO: Should the user decide the positive response or what are the differences?
        return self.positive_responses[response_index].encode(coded_request, **params)

    def encode_negative_response(self, coded_request, response_index=0, **params):
        return self.negative_responses[response_index].encode(coded_request, **params)

    def __call__(self, **params) -> bytes:
        """Encode a request."""
        return self.encode_request(**params)

    def __str__(self):
        return f"DiagService(id={self.id}, semantic={self.semantic})"

    def __repr__(self):
        return self.__str__()

    def __hash__(self) -> int:
        return hash(self.id)

    def __eq__(self, o: object) -> bool:
        return isinstance(o, DiagService) and self.id == o.id


def read_diag_service_from_odx(et_element):

    # logger.info(f"Parsing service based on ET DiagService element: {et_element}")
    short_name = et_element.find("SHORT-NAME").text
    id = et_element.get("ID")

    request_ref_id = et_element.find("REQUEST-REF").get("ID-REF")

    pos_res_ref_ids = [
        el.get("ID-REF") for el in et_element.iterfind("POS-RESPONSE-REFS/POS-RESPONSE-REF")
    ]
    neg_res_ref_ids = [
        el.get("ID-REF") for el in et_element.iterfind("NEG-RESPONSE-REFS/NEG-RESPONSE-REF")
    ]
    functional_class_ref_ids = [
        el.get("ID-REF") for el in et_element.iterfind("FUNCT-CLASS-REFS/FUNCT-CLASS-REF")
    ]
    pre_condition_state_ref_ids = [
        el.get("ID-REF") for el in et_element.iterfind("PRE-CONDITION-STATE-REFS/PRE-CONDITION-STATE-REF")
    ]
    state_transition_ref_ids = [
        el.get("ID-REF") for el in et_element.iterfind("STATE-TRANSITION-REFS/STATE-TRANSITION-REF")
    ]
    long_name = et_element.find(
        "LONG-NAME").text if et_element.find("LONG-NAME") is not None else None
    description = read_description_from_odx(et_element.find("DESC"))
    semantic = et_element.get("SEMANTIC")

    audience = read_audience_from_odx(et_element.find(
        "AUDIENCE")) if et_element.find("AUDIENCE") else None

    diag_service = DiagService(id,
                               short_name,
                               request_ref_id,
                               pos_res_ref_ids,
                               neg_res_ref_ids,
                               long_name=long_name,
                               description=description,
                               semantic=semantic,
                               audience=audience,
                               functional_class_refs=functional_class_ref_ids,
                               pre_condition_state_refs=pre_condition_state_ref_ids,
                               state_transition_refs=state_transition_ref_ids)
    return diag_service



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from dataclasses import dataclass, field
from itertools import chain
from typing import Optional

from .nameditemlist import NamedItemList
from .dataobjectproperty import DataObjectProperty, DopBase, DtcDop, read_data_object_property_from_odx
from .endofpdufield import EndOfPduField, read_end_of_pdu_field_from_odx
from .structures import Structure, read_structure_from_odx
from .units import read_unit_spec_from_odx, UnitSpec
from .globals import logger


def _construct_named_item_list(iterable):
    return NamedItemList(lambda x: x.short_name, iterable)


@dataclass()
class DiagDataDictionarySpec:
    dtc_dops: NamedItemList[DtcDop] = field(
        default_factory=lambda: _construct_named_item_list([])
    )
    data_object_props: NamedItemList[DataObjectProperty] = field(
        default_factory=lambda: _construct_named_item_list([])
    )
    structures: NamedItemList[Structure] = field(
        default_factory=lambda: _construct_named_item_list([])
    )
    end_of_pdu_fields: NamedItemList[EndOfPduField] = field(
        default_factory=lambda: _construct_named_item_list([])
    )
    unit_spec: Optional[UnitSpec] = None

    def __post_init__(self):
        self._all_data_object_properties = _construct_named_item_list(
            chain(self.data_object_props, self.structures,
                  self.end_of_pdu_fields, self.dtc_dops)
        )

        # The attributes are already initialized as (most likely normal) lists.
        # Convert them to NamedItemLists if necessary
        if not isinstance(self.dtc_dops, NamedItemList):
            self.dtc_dops = _construct_named_item_list(
                self.dtc_dops)

        if not isinstance(self.data_object_props, NamedItemList):
            self.data_object_props = _construct_named_item_list(
                self.data_object_props)

        if not isinstance(self.structures, NamedItemList):
            self.structures = _construct_named_item_list(
                self.structures)

        if not isinstance(self.end_of_pdu_fields, NamedItemList):
            self.end_of_pdu_fields = _construct_named_item_list(
                self.end_of_pdu_fields)

    def _build_id_lookup(self):
        id_lookup = {}
        for obj in chain(self.data_object_props,
                         self.structures,
                         self.end_of_pdu_fields):
            id_lookup[obj.id] = obj

        for obj in self.dtc_dops:
            id_lookup.update(obj._build_id_lookup())

        if self.unit_spec:
            id_lookup.update(self.unit_spec._build_id_lookup())

        return id_lookup

    def _resolve_references(self, parent_dl, id_lookup: dict):
        for dop in chain(
                self.dtc_dops,
                self.data_object_props
        ):
            dop._resolve_references(id_lookup)

        for struct in chain(
                self.structures,
                self.end_of_pdu_fields
        ):
            struct._resolve_references(parent_dl, id_lookup)

        if self.unit_spec:
            self.unit_spec._resolve_references(id_lookup)

    @property
    def all_data_object_properties(self):
        return self._all_data_object_properties


def read_diag_data_dictionary_spec_from_odx(et_element):
    # Parse DOP-BASEs
    data_object_props = [read_data_object_property_from_odx(dop_element)
                         for dop_element in et_element.iterfind("DATA-OBJECT-PROPS/DATA-OBJECT-PROP")]

    structures = [read_structure_from_odx(structure_element)
                  for structure_element in et_element.iterfind("STRUCTURES/STRUCTURE")]

    end_of_pdu_fields = [read_end_of_pdu_field_from_odx(eopf_element)
                         for eopf_element in et_element.iterfind("END-OF-PDU-FIELDS/END-OF-PDU-FIELD")]

    dtc_dops = [read_data_object_property_from_odx(dop_element)
                for dop_element in et_element.iterfind("DTC-DOPS/DTC-DOP")]

    if et_element.find("UNIT-SPEC") is not None:
        unit_spec = read_unit_spec_from_odx(et_element.find("UNIT-SPEC"))
    else:
        unit_spec = None

    # TODO: Parse different specs.. Which of them are needed?
    for (path, name) in [
        ('ENV-DATA-DESCS/ENV-DATA-DESC', 'ENV-DATA-DESCs'),
        ('STATIC-FIELDS', 'static fields'),
        ('DYNAMIC-LENGTH-FIELDS/DYNAMIC-LENGTH-FIELD', 'dynamic length fields'),
        ('DYNAMIC-ENDMARKER-FIELDS/DYNAMIC-ENDMARKER-FIELD',
         'dynamic endmarker fields'),
        ('MUXS/MUX', 'MUXs'),
        ('ENV-DATAS/ENV-DATA', 'ENV-DATAs'),
        ('TABLES/TABLE', 'tables'),
    ]:
        num = len(list(et_element.iterfind(path)))
        if num > 0:
            logger.info(f"Not implemented: Did not parse {num} {name}.")

    return DiagDataDictionarySpec(
        data_object_props=data_object_props,
        structures=structures,
        end_of_pdu_fields=end_of_pdu_fields,
        dtc_dops=dtc_dops,
        unit_spec=unit_spec
    )



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

import re
import markdownify

from odxtools import DiagService


def format_desc(desc, ident=0):
    # Collapse whitespaces
    desc = re.sub(r'\s+', ' ', desc)
    # Covert XHTML to Markdown
    desc = markdownify.markdownify(desc)
    # Collapse blank lines
    desc = re.sub(r'(\n\s*)+\n+', '\n', desc).strip()

    if '\n' in desc:
        desc = '\n' + ident * ' ' + \
               ('\n' + ident * ' ').join(desc.split('\n'))
    return desc


def print_diagnostic_service(service: DiagService, print_params=False, print_pre_condition_states=False,
                             print_state_transitions=False, print_audiences=False, allow_unknown_bit_lengths=False):

    print(f" {service.short_name} <ID: {service.id}>")

    if service.description:
        desc = format_desc(service.description, ident=3)
        print(f"  Service description: " + desc)

    if print_pre_condition_states and len(service.pre_condition_states) > 0:
        pre_condition_states_short_names = [pre_condition_state.short_name for pre_condition_state in service.pre_condition_states]
        print(f"  Pre-Condition-States: {', '.join(pre_condition_states_short_names)}")

    if print_state_transitions and len(service.state_transitions) > 0:
        state_transitions = [f"{state_transition.source_short_name} -> {state_transition.target_short_name}" for state_transition in service.state_transitions]
        print(f"  State-Transitions: {', '.join(state_transitions)}")

    if print_audiences and service.audience:
        enabled_audiences_short_names = [enabled_audience.short_name for enabled_audience in service.audience.enabled_audiences]
        print(f"  Enabled-Audiences: {', '.join(enabled_audiences_short_names)}")

    if print_params:
        assert service.request is not None
        assert service.positive_responses is not None
        assert service.negative_responses is not None

        print(f"  Message format of a request:")
        service.request.print_message_format(
            indent=3,
            allow_unknown_lengths=allow_unknown_bit_lengths)

        print(
            f"  Number of positive responses: {len(service.positive_responses)}")
        if len(service.positive_responses) == 1:
            assert service.positive_responses[0] is not None

            print(f"  Message format of a positive response:")
            service.positive_responses[0].print_message_format(
                indent=3,
                allow_unknown_lengths=allow_unknown_bit_lengths)

        print(
            f"  Number of negative responses: {len(service.negative_responses)}")
        if len(service.negative_responses) == 1:
            assert service.negative_responses[0] is not None

            print(f"  Message format of a negative response:")
            service.negative_responses[0].print_message_format(
                indent=3,
                allow_unknown_lengths=allow_unknown_bit_lengths)

    if ((service.positive_responses and len(service.positive_responses) > 1)
            or (service.negative_responses and len(service.negative_responses) > 1)):
        # Does this ever happen?
        raise NotImplementedError(
            f"The diagnostic service {service.id} offers more than one response!")
